/* Device Driver for XADIO Cards  */
/* Copyright (1999-2003) Brueckner & Jarosch Erfurt Germany */

#undef CONFIG_DEBUG_IOVIRT
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/ctype.h>
#include <linux/interrupt.h>

#include "xadio.h"
#define USE_RPM_CODE    /* FPGA misst Zeit zwischen 4 Inkrementzyklen und schreibt Zaehlerergebnis int fill1, 50Mhz_Takte/Inkrement=fill1*2*(div+1) */
#define MINOR_MASTER -1 /* disabled: -1 , Wenn Slave Karten ueber externe Leitung den 50MHz Takt bekommen sollen, muss MINOR_MASTER >=0 gesetzt werden */
                        /* dann wird bei SLAVE Karten ein anderes bitfile geladen: xadio_fpgas.h anstatt xadio_fpga.h */
#define noUSE_DEBUG_IN_CONTROL




/* Offset in EEPRom where config starts */
#define XADIO_CONFIGOFFSET 256

/* Maximum Count of Cards (can bee increased to 255) */

/* Revision history
   XADIO_DRIVERVERSION>=90:
   - providing tic counter between increment changes	
   XADIO_DRIVERVERSION>=81:
   - modeword 0x200 needs to be set for analog input negation (XADIO_REV<2.1). This will be done automatically in the driver by reading XADIO_REV from eeprom.
   - modeword 0x18  controls 2 relais for testmode. New xadio command: settestmode <mode> (mode=0 normal,mode=1 0V,mode=3 Umax)
   - dac correction can be disabled by using offset=0 and gain=0 in eeprom data
   XADIO_DRIVERVERSION>=70:
   - setda0/setda1 do a digital dac output correction, correction data will be read from eeprom
   using /etc/modules.conf.local instead of /etc/sysconfig/xadio

    FPGA231 bei BMW usw.
    FPGA232 letzter Stand
    FPGA233 gemischt
*/

#define XADIO_NO                          4
#define XADIO_DRIVERVERSION               106    /* Release Versionen bitte mit ungerader Nummer */
#define XADIO_DEFAULT_BRECORDS            1024
#define XADIO_DEFAULT_BCOUNT              8
#define XADIO_DEFAULT_NOSUBVENDORCHECK    0
#define XADIO_DEFAULT_HWVERSION           24	/* 
						1
						2 = Leiterplatte 2.0 FPGA-CORE=2.5V,Trimmer             
						21= Leiterplatte 2.1 FPGA-CORE=2.5V,FestWid,Relais  	(R15,R16 Originalbestueckung des Jahres 2005)
						22= Leiterplatte 2.0 FPGA-CORE=1.8V,Trimmer		        (R15,R16 nachtraeglich getauscht)
						23= Leiterplatte 2.1 FPGA-CORE=1.8V,FestWid,Relais  	(R15,R16 Originalbestueckung ab Jahr 2007 oder mit getauschtem R15,R16 des Jahres 2005)
						24= Leiterplatte 2.4 Onboard kalibrierung nun benutzbar, kleinerer(Leistung) DC/DC Wandler
						*/

#ifndef PCI_DEVICE_ID_XADIO
#define PCI_DEVICE_ID_XADIO 0xadda
#endif
#ifndef PCI_VENDOR_ID_DRESEARCH
#define PCI_VENDOR_ID_DRESEARCH 0x15eb
#endif
#ifndef PCI_DEVICE_ID_9054
#define PCI_DEVICE_ID_9054 0x9054
#endif

#ifndef PCI_ANY_ID
#define PCI_ANY_ID (~0)
#endif

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/pgtable.h>

#define ARRSIZE(a)          ((int)(sizeof(a)/sizeof(a[0])))
#define OFFSET(type,member) ((int)(&((type*)0)->member))
#define ARRPOS(ptr,arr)     ((ptr)-(arr))
#define MIN(a,b)            (((a)<(b))?(a):(b))
#define MAX(a,b)            (((a)>(b))?(a):(b))
#define LO(ll)              ((unsigned int)((ll      )&0xffffffffLL))
#define HI(ll)              ((unsigned int)(((ll)>>32)&0xffffffffLL))
#define HILO(ll)            HI(ll),LO(ll)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#ifndef init_waitqueue_head
#define init_waitqueue_head(x)
#endif
#ifndef pci_resource_len
static int pci_resource_len_fkt(struct pci_dev *pdev,int region) {
u32 mask,b2;
  int offset=0; 
  switch(region) {
    case 0: offset=PCI_BASE_ADDRESS_0; break;
    case 1: offset=PCI_BASE_ADDRESS_1; break;
    case 2: offset=PCI_BASE_ADDRESS_2; break;
    case 3: offset=PCI_BASE_ADDRESS_3; break;
  }
  pci_read_config_dword(pdev,offset,&b2);
  pci_write_config_dword(pdev,offset,0xffffffff);
  pci_read_config_dword(pdev,offset,&mask);
  pci_write_config_dword(pdev,offset,b2);
  return ((mask&PCI_BASE_ADDRESS_MEM_MASK)^0xffffffff)+1;
}
#define pci_resource_len pci_resource_len_fkt
#endif
#define pci_request_regions(x,y) (0)
#define pci_release_regions(x)
#define pci_enable_device(x) (0)
#define pci_disable_device(x)
#define pci_resource_start(pdev,i) ((pdev->base_address[i])& PCI_BASE_ADDRESS_MEM_MASK)
#ifndef module_init
#define module_init(a) int init_module(void) { return a(); }
#endif
#ifndef module_exit
#define module_exit(a) void cleanup_module(void) { a(); }
#endif
#ifndef MODULE_DEVICE_TABLE
#define MODULE_DEVICE_TABLE(x,y)
#endif
#ifndef __devexit
#define __devexit
#endif
#ifndef __devinit
#define __devinit
#endif

#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#undef MOD_INC_USE_COUNT
#define MOD_INC_USE_COUNT
#undef MOD_DEC_USE_COUNT
#define MOD_DEC_USE_COUNT
#else
#ifndef IRQ_RETVAL
typedef void irqreturn_t;
#define IRQ_RETVAL(x)
#endif
#endif

#ifndef MODULE_LICENSE
#define MODULE_LICENSE(x)
#endif

#ifndef  SEEK_SET
 #define SEEK_SET 0
 #define SEEK_CUR 1
 #define SEEK_END 2
#endif	

#define MAX_BUFFER 128

// Module Parameter 
static int debug = 0; 
static int polling_interval = 1000; /* Polling in msec for xadio_read*/
static int major = 0x46; 
static int brecords = XADIO_DEFAULT_BRECORDS;
static int bcount = XADIO_DEFAULT_BCOUNT;
static int nosubvendorcheck = XADIO_DEFAULT_NOSUBVENDORCHECK;
static int hwversion = XADIO_DEFAULT_HWVERSION;
static int config_rom_error[XADIO_NO];
// Globals
static int bsize = 0; 

// Structures 

typedef struct {
  u32 dst;
  u32 src;  /* ist immer konstant bei allen Eintraegen */
  u32 size; /* ist immer konstant bei allen Eintraegen */
  u32 next; /* Ringverkettung, fordert DMA */
} xadio_dma_desc;

typedef struct {
  unsigned int   shift;                   /* 1byte in eeprom */
  int            offsets[XADIO_CHANNELS]; /* 8byte in eeprom  */
} xadio_offset9_t;

typedef struct {
  unsigned int    frequency;     /* 4byte (3 wuerden reichen) in eeprom */
  xadio_offset9_t  corr0;        /* 9byte in eeprom */
  xadio_offset9_t  corr1;        /* 9byte in eeprom */
} xadio_offset_t;    /* 22 byte im EEPROM */

typedef struct {
  int   offs0;                  /* digit->digit */
  int   offs1;                  /* */
  int   xm;                     /* */
  int   uend;                   /* digit->u */
  int  dil;
} xadio_adccorr_t;
/* adccorr = (adc-offs0-xm) - (adc-offs0-xm)*offs1/0x8000 + xm; */


#define XADIO_EE_MAX_ENTRIES        10

typedef struct {
  unsigned int   version;                      /* 1byte in eeprom */
  unsigned int   hwversion;                    /* 1byte in eeprom */
  unsigned int   cardyear;                     /* 1byte in eeprom */
  unsigned int   cardnumber;                   /* 3byte in eeprom */
  unsigned int   jumpers[XADIO_CHANNELS];      /* 4byte(not 8!) in eeprom */
  unsigned int   entries;                      /* 1byte in eeprom */
  xadio_offset_t entry[XADIO_EE_MAX_ENTRIES];  /* 10*22 byte */
  int            dac_m[2];                     /* dac: 6 byte in eeprom */
  int            dac_y0[2]; 
  /* not in eeprom */
  xadio_adccorr_t adccorr[XADIO_CHANNELS];     /* depends on the current sampling frequ, set by xadio_registerfreq */
  unsigned int   curr_freq;
} xadio_eepromio_t;

typedef struct {
  int flags;
  char name[10];
  struct pci_dev *pdev;
  volatile unsigned char *base0;
  volatile unsigned char *base2;
  int size0;
  int size2;
  int size3;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
  wait_queue_head_t dma_wait;
  wait_queue_head_t local_wait;
  wait_queue_head_t write_wait;
#else
  struct wait_queue *dma_wait;
  struct wait_queue *local_wait;
  struct wait_queue *write_wait;
#endif
  xadio_dma_desc *darr[MAX_BUFFER]; /* Ring */
  unsigned char * barr[MAX_BUFFER]; /* deswegen auch hier ein Ring, Ringzaehlerspeicher */
  int aktbuf;                                                 
  int counter0;
  int counter1;
  void *drvdata;  /* used for kernel<2.4 */
  char configrom[XADIO_CONFIG_ROMSIZE];
#ifdef USE_DEBUG_IN_CONTROL
  char debugtext[256]; /* __mo__ debugging */
  int  debug_interrupts;
#endif
  xadio_eepromio_t eeprom;
  int   glFpgaLoaded;
} xadio_t;

static xadio_t xadio_table[XADIO_NO];
static int xadioencodersize[XADIO_NO][2]={{1024,1024},{1024,1024}};
static int ResultNINIT[XADIO_NO];

#define XADIO_VALID_BIT_POS      0
#define XADIO_USED_BIT_POS       1
#define XADIO_DMA_RUNNING_POS    2

#define XADIO_VALID_MASK         (1<<XADIO_VALID_BIT_POS)
#define XADIO_USED_MASK          (1<<XADIO_USED_BIT_POS)
#define XADIO_DMA_MASK           (1<<XADIO_DMA_RUNNING_POS)

#define XADIO_NAME(minor)        (xadio_table[minor].name)
#define XADIO_VALID(minor)       (minor<XADIO_NO && (xadio_table[minor].flags & XADIO_VALID_MASK))
#define XADIO_USED(minor)        (minor<XADIO_NO && (xadio_table[minor].flags & XADIO_USED_MASK))

static void xadio_setdacuv(int minor,int channel,int value);
static void xadio_setmodeword(int minor,unsigned int modeword);
static unsigned int xadio_getmodeword(int minor); 
static int xadio_getfpgaversion(int minor,int channel);
int xadio_get_dac_corr(int minor,int dac,int uV);
#if (MINOR_MASTER>=0)
static void xadio_master(int minor);
#endif
static void xadio_greenled(int minor,int onoff); 
static void xadio_redled  (int minor,int onoff); 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
void pci_set_drvdata(struct pci_dev *pdev,void *drvdata) {
int i;
int minor=((int)drvdata)-1;
  for(i=0;i<ARRSIZE(xadio_table);i++) {
    if(xadio_table[i].pdev==pdev) {
      xadio_table[i].drvdata=drvdata;
      return;
    }
  }
  printk(KERN_WARNING "%s no slot in pci_set_drvdata\n",XADIO_NAME(minor));
}

void *pci_get_drvdata(struct pci_dev *pdev) {
int i;
  for(i=0;i<ARRSIZE(xadio_table);i++) {
    if(xadio_table[i].pdev==pdev) return xadio_table[i].drvdata;
  }
  printk(KERN_WARNING "%s no such pdev in pci_get_drvdata\n",XADIO_STRING);
  return 0;
}

struct pci_device_id;

#endif

#define RFL(minor,offset)        readl(      xadio_table[minor].base2+offset)
#define WFL(minor,offset,value) writel(value,xadio_table[minor].base2+offset)
#define RFB(minor,offset)        readb(      xadio_table[minor].base2+offset)
#define WFB(minor,offset,value) writeb(value,xadio_table[minor].base2+offset)
#define RPL(minor,offset)        readl(      xadio_table[minor].base0+offset)
#define WPL(minor,offset,value) writel(value,xadio_table[minor].base0+offset)
#define RPW(minor,offset)        readw(      xadio_table[minor].base0+offset)
#define WPW(minor,offset,value) writew(value,xadio_table[minor].base0+offset)
#define RPB(minor,offset)        readb(      xadio_table[minor].base0+offset)
#define WPB(minor,offset,value) writeb(value,xadio_table[minor].base0+offset)

#define GetPlx_las0rr(minor)         RPL(minor,0x00)
#define GetPlx_las0ba(minor)         RPL(minor,0x04)
#define GetPlx_marbr(minor)          RPL(minor,0x08)
#define GetPlx_lbrd0(minor)          RPL(minor,0x18)
#define GetPlx_intcsr(minor)         RPL(minor,0x68)
#define GetPlx_cntrl(minor)          RPL(minor,0x6c)
#define GetPlx_dmamode0(minor)       RPL(minor,0x80)
#define GetPlx_dmapadr0(minor)       RPL(minor,0x84)
#define GetPlx_dmaladr0(minor)       RPL(minor,0x88)
#define GetPlx_dmasiz0(minor)        RPL(minor,0x8c)
#define GetPlx_dmadpr0(minor)        RPL(minor,0x90)
#define GetPlx_dmathr(minor)         RPL(minor,0xb0)
#define GetPlx_las1rr(minor)         RPL(minor,0xf0)
#define GetPlx_las1ba(minor)         RPL(minor,0xf4)
#define GetPlx_lbrd1(minor)          RPL(minor,0xf8)
#define GetPlx_prot(minor)           RPW(minor,0x0e)
#define GetPlx_dmacsr0(minor)        RPB(minor,0xa8)
#define GetPlx_dmacsr1(minor)        RPB(minor,0xa9)
#define SetPlx_las0rr(minor,value)   WPL(minor,0x00,value)
#define SetPlx_las0ba(minor,value)   WPL(minor,0x04,value)
#define SetPlx_marbr(minor,value)    WPL(minor,0x08,value)
#define SetPlx_lbrd0(minor,value)    WPL(minor,0x18,value)
#define SetPlx_intcsr(minor,value)   WPL(minor,0x68,value)
#define SetPlx_cntrl(minor,value)    WPL(minor,0x6c,value)
#define SetPlx_dmamode0(minor,value) WPL(minor,0x80,value)
#define SetPlx_dmapadr0(minor,value) WPL(minor,0x84,value)
#define SetPlx_dmaladr0(minor,value) WPL(minor,0x88,value)
#define SetPlx_dmasiz0(minor,value)  WPL(minor,0x8c,value)
#define SetPlx_dmadpr0(minor,value)  WPL(minor,0x90,value)
#define SetPlx_dmathr(minor,value)   WPL(minor,0xb0,value)
#define SetPlx_las1rr(minor,value)   WPL(minor,0xf0,value)
#define SetPlx_las1ba(minor,value)   WPL(minor,0xf4,value)
#define SetPlx_lbrd1(minor,value)    WPL(minor,0xf8,value)
#define SetPlx_prot(minor,value)     WPW(minor,0x0e,value)
#define SetPlx_dmacsr0(minor,value)  WPB(minor,0xa8,value)
#define SetPlx_dmacsr1(minor,value)  WPB(minor,0xa9,value)
#define GetFpga_adu(minor,channel)   RFL(minor,0x00+(channel<<2))
#define SetFpga_dac(minor,channel,value)   WFL(minor,0x28+(channel<<2),value)
#define GetFpga_dac(minor,channel)   RFL(minor,0x28+(channel<<2))
#define GetFpga_digio(minor)         RFL(minor,0x20)
#define GetFpga_inc0(minor)          RFL(minor,0x24)

  /* 1.June2005 212 */
  #define GetFpga_timer64(minor)       RFL(minor,0x3C)
  #define GetFpga_timer(minor)         RFL(minor,0x30)
  #define GetFpga_inc1(minor)          RFL(minor,0x34)
  #define GetFpga_fill1(minor)         RFL(minor,0x38)
  #define GetFpga_fill2(minor)         RFL(minor,0x3c)
#define GetFpga_ien(minor)           RFL(minor,0x44)
#define GetFpga_isrc(minor)          RFL(minor,0x48)
#define GetFpga_cntrl(minor)         RFL(minor,0x4c)
#define SetFpga_digio(minor,value)   WFL(minor,0x20,value)
#define SetFpga_timer(minor,value)   WFL(minor,0x30,value)
#define SetFpga_ien(minor,value)     WFL(minor,0x44,value)
#define SetFpga_isrc(minor,value)    WFL(minor,0x48,value)
#define SetFpga_cntrl(minor,value)   WFL(minor,0x4c,value)

#define SetFpga_div0(minor,value) 	WFL(minor,0x50,value)
#define SetFpga_div1(minor,value) 	WFL(minor,0x54,value)
#define GetFpga_div0(minor) 		RFL(minor,0x50)
#define GetFpga_div1(minor) 		RFL(minor,0x54)
#define GetFpga_version(minor)          RFL(minor,0x58)


static void xadio_delay(void) {
  __udelay(1);
  wmb();
}

static void xadio_sleep(int minor,int msec) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
struct wait_queue *sleep_q=0;
interruptible_sleep_on_timeout(&sleep_q,(msec*HZ)/1000);
#else
wait_queue_head_t wqh;
  init_waitqueue_head(&wqh);
  interruptible_sleep_on_timeout(&wqh,(msec*HZ)/1000);
#endif
}

#ifdef CONFIG_DEBUG_SLAB
static int xadio_roundmemsize(int size,int align) {
  return size+align;
}

static void* xadio_roundptr(void* ptr,int align) {
  return (void*)(((long)ptr+align-1)&(~(align-1)));
}

static int xadio_roundmemsize16(int size) {
  return xadio_roundmemsize(size,16);
}

static void* xadio_roundptr16(void* ptr) {
  return xadio_roundptr(ptr,16);
}
#else
static int xadio_roundmemsize16(int size) { return size; }
static void* xadio_roundptr16(void* ptr) { return ptr; }
#endif


/* ------------------------- EEPROM-STUFF----------------------------- */
static int xadio_value2uv(int minor,int channel,int value);
#define BYTESPERBYTE    2
#define EEPROMTRANSFERSIZE XADIO_CONFIG_ROMSIZE*BYTESPERBYTE

/* default EEPROM values */
#define XADIO_EE_DEFAULT_VERSION    4
#define XADIO_EE_DEFAULT_HWVERSION  2
/* +/- 10V */
#define XADIO_EE_DEFAULT_JUMPER     9

#define XADIO_EE_OFFSET_VERSION     0
#define XADIO_EE_SIZEOF_VERSION     1
#define XADIO_EE_OFFSET_HWVERSION   (XADIO_EE_OFFSET_VERSION+XADIO_EE_SIZEOF_VERSION)
#define XADIO_EE_SIZEOF_HWVERSION   1
#define XADIO_EE_OFFSET_CARDNUMBER  (XADIO_EE_OFFSET_HWVERSION+XADIO_EE_SIZEOF_HWVERSION)
#define XADIO_EE_SIZEOF_CARDNUMBER  3
#define XADIO_EE_OFFSET_CARDYEAR    (XADIO_EE_OFFSET_CARDNUMBER+XADIO_EE_SIZEOF_CARDNUMBER)
#define XADIO_EE_SIZEOF_CARDYEAR    1
#define XADIO_EE_OFFSET_ENTRIES     (XADIO_EE_OFFSET_CARDYEAR+XADIO_EE_SIZEOF_CARDYEAR)
#define XADIO_EE_SIZEOF_ENTRIES     1
#define XADIO_EE_OFFSET_JUMPERS     (XADIO_EE_OFFSET_ENTRIES+XADIO_EE_SIZEOF_ENTRIES)
#define XADIO_EE_SIZEOF_JUMPERS     XADIO_CHANNELS/2

#define XADIO_EE_OFFSET_CALIBTEMP   (XADIO_EE_OFFSET_JUMPERS+XADIO_EE_SIZEOF_JUMPERS)
#define XADIO_EE_SIZEOF_CALIBTEMP   1

#define XADIO_EE_OFFSET_DRIFT0      (XADIO_EE_OFFSET_CALIBTEMP+XADIO_EE_SIZEOF_CALIBTEMP)
#define XADIO_EE_SIZEOF_DRIFT0      8

#define XADIO_EE_OFFSET_DRIFT2      (XADIO_EE_OFFSET_DRIFT0+XADIO_EE_SIZEOF_DRIFT0)
#define XADIO_EE_SIZEOF_DRIFT2      8

#define XADIO_EE_OFFSET_DAC0        (XADIO_EE_OFFSET_DRIFT2+XADIO_EE_SIZEOF_DRIFT2)
#define XADIO_EE_SIZEOF_DAC0        3

#define XADIO_EE_OFFSET_DAC1        (XADIO_EE_OFFSET_DAC0+XADIO_EE_SIZEOF_DAC0)
#define XADIO_EE_SIZEOF_DAC1        3

#define XADIO_EE_OFFSET_FILLER      (XADIO_EE_OFFSET_JUMPERS+XADIO_EE_SIZEOF_JUMPERS)
#define XADIO_EE_OFFSET_OFFSETS     34
#define XADIO_EE_OFFSET_COUNT       10
#define XADIO_EE_SIZEOF_OFFSETS     220
#define XADIO_EE_SIZEOF_SHIFT       1
#define XADIO_EE_SIZEOF_FREQUENCY   4
#define XADIO_EE_SIZEOF_OFFSET      1
#define XADIO_EE_SIZEOF_CHECKSUM    2
#define XADIO_EE_OFFSET_CHECKSUM    (XADIO_CONFIG_ROMSIZE-XADIO_EE_SIZEOF_CHECKSUM)

#ifdef LATER
typedef struct {
  const char *name;
  int offset;
  int len;
  int type;
  int channel;
  int frequency;
} Xadio_EEpromLayoutItem;

static Xadio_EEpromLayoutItem eelayout[] = {
  { "EEVersion"  , 0, 1, 1, 0, 0 },
  { "EEHWVersion", 1, 1, 1, 0, 0 },
  { "EECardNum",   2, 3, 1, 0, 0 },
  { "EECardYear",  5, 1, 1, 0, 0 },
  { "EEEntries",   6, 1, 1, 0, 0 },
  { "EEJumper0",   7, 4, 2, 0, 0 },
  { "EEJumper1",   7, 4, 2, 1, 0 },
  { "EEJumper2",   7, 4, 2, 2, 0 },
  { "EEJumper3",   7, 4, 2, 3, 0 },
  { "EEJumper4",   7, 4, 2, 4, 0 },
  { "EEJumper5",   7, 4, 2, 5, 0 },
  { "EEJumper6",   7, 4, 2, 6, 0 },
  { "EEJumper7",   7, 4, 2, 7, 0 },
  { "EEFiller",   11,23, 0, 0, 0 },
#define SOFF 34
  { "EEF0"   ,    SOFF+ 0, 4, 1, 0, 0 },
  { "EEF0S0",     SOFF+ 4, 1, 1, 0, 0 },
  { "EEF0C0O0",   SOFF+ 5, 1, 0, 0, 0 },
  { "EEF0C1O0",   SOFF+ 6, 1, 0, 0, 0 },
  { "EEF0C2O0",   SOFF+ 7, 1, 0, 0, 0 },
  { "EEF0C3O0",   SOFF+ 8, 1, 0, 0, 0 },
  { "EEF0C4O0",   SOFF+ 9, 1, 0, 0, 0 },
  { "EEF0C5O0",   SOFF+10, 1, 0, 0, 0 },
  { "EEF0C6O0",   SOFF+11, 1, 0, 0, 0 },
  { "EEF0C7O0",   SOFF+12, 1, 0, 0, 0 },
  { "EEF0S1",     SOFF+13, 1, 1, 0, 0 },
  { "EEF0C0O1",   SOFF+14, 1, 0, 0, 0 },
  { "EEF0C1O1",   SOFF+15, 1, 0, 0, 0 },
  { "EEF0C2O1",   SOFF+16, 1, 0, 0, 0 },
  { "EEF0C3O1",   SOFF+17, 1, 0, 0, 0 },
  { "EEF0C4O1",   SOFF+18, 1, 0, 0, 0 },
  { "EEF0C5O1",   SOFF+19, 1, 0, 0, 0 },
  { "EEF0C6O1",   SOFF+20, 1, 0, 0, 0 },
  { "EEF0C7O1",   SOFF+21, 1, 0, 0, 0 },
#undef SOFF
};

#endif

static int xadio_eepromgetint(int minor,int *result,int offset,int len) {
const char *eeprom=xadio_table[minor].configrom;
int i,InvBit=0,mask=0;
  if(offset<0 || len<0 || len>((int)sizeof(int)) || offset+len>XADIO_CONFIG_ROMSIZE){
     printk(KERN_WARNING "%s Error in xadio_eepromgetint: offset:%d len:%d\n",XADIO_NAME(minor),offset,len);
     return -ERANGE;
  }
  if(result) *result=0;
  for(i=offset+len-1; i>=offset; i--,mask+=8){
     if(result) {
        (*result) = (*result) + ((((unsigned)eeprom[i])&255)<<mask);
     }
     InvBit = eeprom[i] & 0x80;
  }
  if(result && InvBit){
     if(len==1) *result |= 0xffffff00;
     if(len==2) *result |= 0xffff0000;
     if(len==3) *result |= 0xff000000;
  }
  return 0;
}

static int xadio_eepromgetuint(int minor,unsigned int *result,int offset,int len) {
const char *eeprom=xadio_table[minor].configrom;
int i,mask=0;
  if(!result) return -ERANGE;
  if(offset<0 || len<0 || len>((int)sizeof(int)) || offset+len>XADIO_CONFIG_ROMSIZE){
     printk(KERN_WARNING "%s Error in xadio_eepromgetuint: offset:%d len:%d\n",XADIO_NAME(minor),offset,len);
     return -ERANGE;
  }
  *result=0;
  for(i=offset+len-1; i>=offset; i--,mask+=8){
    (*result) = (*result) + ((((unsigned)eeprom[i])&255)<<mask);
  }
  return 0;
}

static int  xadio_eepromgetnibbleint(int minor,int *pInt,int nibbel_offset,int nibbel_len)
{
const char *eeprom=xadio_table[minor].configrom;
int i,InvBit=0,intpos=0,mask=0;

  if(!pInt) return -ERANGE;
  if(nibbel_offset<0 || nibbel_len<0 || nibbel_offset+nibbel_len>XADIO_CONFIG_ROMSIZE*2){
     printk(KERN_WARNING "%s xadio_eepromgetnibbleint out of range, nibbel_offset=%d nibbel_len=%d\n",XADIO_NAME(minor),nibbel_offset,nibbel_len);
     return -ERANGE;
  }
  *pInt=0;
  for(i=nibbel_offset+nibbel_len-1; i>=nibbel_offset; i--,intpos++,mask+=4){
     char data = eeprom[i/2]>>(i%2==1?0:4);
     (*pInt) = (*pInt) + ((((unsigned)data)&0xf)<<mask);
     InvBit = data & 0x08;
  }
  if(pInt && InvBit){
     if(nibbel_len==1) *pInt |= 0xfffffff0;
     if(nibbel_len==2) *pInt |= 0xffffff00;
     if(nibbel_len==3) *pInt |= 0xfffff000;
     if(nibbel_len==4) *pInt |= 0xffff0000;
     if(nibbel_len==5) *pInt |= 0xfff00000;
     if(nibbel_len==6) *pInt |= 0xff000000;
     if(nibbel_len==7) *pInt |= 0xf0000000;
  }
  return 0;
}

static int xadio_eepromgetnibble(int minor,unsigned int *result,int offset,int pos) {
const char *eeprom=xadio_table[minor].configrom;
  if(!result) return -ERANGE;
  if(offset<0 || pos<0 || pos>XADIO_CHANNELS || offset+pos/2>XADIO_CONFIG_ROMSIZE){
     printk(KERN_WARNING "%s Error in xadio_eepromgetnibble: offset:%d pos:%d\n",XADIO_NAME(minor),offset,pos);
     return -ERANGE;
  }
  *result=eeprom[offset+pos/2];
  if(pos&1) {
    *result&=0xf;
  } else {
    *result=(*result)>>4 & 0xf;
  }
  return 0;
}

static int xadio_eepromgetoffsets(int minor) {
xadio_offset_t prev;
int f,ch;
  int pos=XADIO_EE_OFFSET_OFFSETS;
  memset(&prev,0,sizeof(prev));
  for(f=0;f<(int)xadio_table[minor].eeprom.entries;f++) {
    xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.entry[f].frequency),pos,XADIO_EE_SIZEOF_FREQUENCY);
    pos+=XADIO_EE_SIZEOF_FREQUENCY;
    xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.entry[f].corr0.shift),pos,XADIO_EE_SIZEOF_SHIFT);
    pos+=XADIO_EE_SIZEOF_SHIFT;
    for(ch=0;ch<XADIO_CHANNELS;ch++){
      xadio_eepromgetint(minor,&(xadio_table[minor].eeprom.entry[f].corr0.offsets[ch]),pos,XADIO_EE_SIZEOF_OFFSET);
      pos+=XADIO_EE_SIZEOF_OFFSET;
      xadio_table[minor].eeprom.entry[f].corr0.offsets[ch]=
       prev.corr0.offsets[ch]=
        (xadio_table[minor].eeprom.entry[f].corr0.offsets[ch]<<
         xadio_table[minor].eeprom.entry[f].corr0.shift)+
        prev.corr0.offsets[ch];
    }
    xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.entry[f].corr1.shift),pos,XADIO_EE_SIZEOF_SHIFT);
    pos+=XADIO_EE_SIZEOF_SHIFT;
    for(ch=0;ch<XADIO_CHANNELS;ch++){
      xadio_eepromgetint(minor,&(xadio_table[minor].eeprom.entry[f].corr1.offsets[ch]),pos,XADIO_EE_SIZEOF_OFFSET);
      pos+=XADIO_EE_SIZEOF_OFFSET;
      xadio_table[minor].eeprom.entry[f].corr1.offsets[ch]=
       prev.corr1.offsets[ch]=
        (xadio_table[minor].eeprom.entry[f].corr1.offsets[ch]<<
         xadio_table[minor].eeprom.entry[f].corr1.shift)+
        prev.corr1.offsets[ch];
    }
  }
  return 0;
}

static int xadio_eepromcalcchecksum(int minor,unsigned int *result) {
int i;
  (*result)=0;
  for(i=0;i<XADIO_EE_OFFSET_CHECKSUM;i++) {
    (*result)+=(unsigned char)(xadio_table[minor].configrom[i]);
  }
  return 0;
}

int GetDacCorr(int HardwareRevision)
{
    if(HardwareRevision<=2 || HardwareRevision==22)
        return 0x10000;	// Trimmer
    return 62000;	// Festwiderstaende erzeugen 12V anstatt 10V	
}

static int xadio_eepromimport(int minor) {
unsigned int RealCheckSum;
unsigned int CheckSum;
int  j;

   xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.version),XADIO_EE_OFFSET_VERSION,XADIO_EE_SIZEOF_VERSION);
   if(xadio_table[minor].eeprom.version<=0 ||
      xadio_table[minor].eeprom.version>=127 ||
      (xadio_table[minor].eeprom.version!=XADIO_EE_DEFAULT_VERSION && xadio_table[minor].eeprom.version!=2 && xadio_table[minor].eeprom.version!=3 && xadio_table[minor].eeprom.version !=1) 
     ){
     printk(KERN_WARNING "%s EEprom version(=%d) mismatch, %d expected!\n",XADIO_NAME(minor),xadio_table[minor].eeprom.version,XADIO_EE_DEFAULT_VERSION);
     return -EIO;
   }

   if(xadio_table[minor].eeprom.version==1){
     printk(KERN_WARNING "%s EEprom version(%d) mismatch, %d expected! Assuming +-10V\n",XADIO_NAME(minor),xadio_table[minor].eeprom.version,XADIO_EE_DEFAULT_VERSION);
   }

   xadio_eepromgetuint(minor,&CheckSum,XADIO_EE_OFFSET_CHECKSUM,XADIO_EE_SIZEOF_CHECKSUM);
   xadio_eepromcalcchecksum(minor,&RealCheckSum);
   if(RealCheckSum!=CheckSum){
     xadio_greenled(minor,0);
     xadio_redled(minor,1);
     printk(KERN_WARNING "%s wrong checksum %d!=%d(expected)",XADIO_NAME(minor),RealCheckSum,CheckSum);
     return -EIO;
   }
   xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.hwversion),XADIO_EE_OFFSET_HWVERSION,XADIO_EE_SIZEOF_HWVERSION);
   xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.cardyear),XADIO_EE_OFFSET_CARDYEAR,XADIO_EE_SIZEOF_CARDYEAR);
   xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.cardnumber),XADIO_EE_OFFSET_CARDNUMBER,XADIO_EE_SIZEOF_CARDNUMBER);
   /* dac's */
   if(xadio_table[minor].eeprom.version>=4){
     xadio_eepromgetnibbleint (minor,&(xadio_table[minor].eeprom.dac_m[0]) ,XADIO_EE_OFFSET_DAC0*2+0,3);
     xadio_eepromgetnibbleint (minor,&(xadio_table[minor].eeprom.dac_y0[0]),XADIO_EE_OFFSET_DAC0*2+3,3);
     xadio_eepromgetnibbleint (minor,&(xadio_table[minor].eeprom.dac_m[1]) ,XADIO_EE_OFFSET_DAC1*2+0,3);
     xadio_eepromgetnibbleint (minor,&(xadio_table[minor].eeprom.dac_y0[1]),XADIO_EE_OFFSET_DAC1*2+3,3);
     if(xadio_table[minor].eeprom.dac_m[0]!=0) xadio_table[minor].eeprom.dac_m[0]+=GetDacCorr(xadio_table[minor].eeprom.hwversion);
     if(xadio_table[minor].eeprom.dac_m[1]!=0) xadio_table[minor].eeprom.dac_m[1]+=GetDacCorr(xadio_table[minor].eeprom.hwversion);
   }
  
   if(debug>0) 
      printk(KERN_WARNING "%s dac %d,%d %d,%d\n",XADIO_NAME(minor),xadio_table[minor].eeprom.dac_m[0]/2-0x8000,xadio_table[minor].eeprom.dac_y0[0],xadio_table[minor].eeprom.dac_m[1]/2-0x8000,xadio_table[minor].eeprom.dac_y0[1]);
    
   /* */
   xadio_eepromgetuint(minor,&(xadio_table[minor].eeprom.entries),XADIO_EE_OFFSET_ENTRIES,XADIO_EE_SIZEOF_ENTRIES);
   for(j=0;j<XADIO_CHANNELS;j++){
    xadio_eepromgetnibble(minor,&(xadio_table[minor].eeprom.jumpers[j]),XADIO_EE_OFFSET_JUMPERS,j);
     if(xadio_table[minor].eeprom.version==1){
       xadio_table[minor].eeprom.jumpers[j]=XADIO_EE_DEFAULT_JUMPER;
     }
   }
   if(xadio_table[minor].eeprom.entries>XADIO_EE_MAX_ENTRIES){
      printk(KERN_WARNING "%s Too many offsets(%d), max. %d expected!",XADIO_NAME(minor),xadio_table[minor].eeprom.entries,XADIO_EE_MAX_ENTRIES);
      xadio_table[minor].eeprom.entries=XADIO_EE_MAX_ENTRIES;
   }
   xadio_eepromgetoffsets(minor);
   xadio_table[minor].eeprom.version=XADIO_EE_DEFAULT_VERSION; /* upgrade */
   /*if(debug>0)*/
      printk(KERN_INFO "%s xadio_eepromimport ok,checksum=%08x hwversion=%d cardyear=%d ResultNINIT=%d.\n",XADIO_NAME(minor),RealCheckSum,xadio_table[minor].eeprom.hwversion,xadio_table[minor].eeprom.cardyear,ResultNINIT[minor]);
   if(xadio_table[minor].eeprom.hwversion>=22){
     /* 22= Leiterplatte 2.0 FPGA-CORE=1.8V,Trimmer (R15,R16 nachtraeglich getauscht)
        23= Leiterplatte 2.1 FPGA-CORE=1.8V,Relais (R15,R16 Originalbestueckung ab Jahr 2007 oder mit getauschtem R15,R16 des Jahres 2005) */
     if(ResultNINIT[minor]==0)
	printk(KERN_WARNING "%s hwversion=%d cardyear=%d  error: ResultNINIT=%d\n",XADIO_NAME(minor),xadio_table[minor].eeprom.hwversion,xadio_table[minor].eeprom.cardyear,ResultNINIT[minor]);
     else	
  	xadio_redled(minor,0);
   }
   else{ // war falsche Spannung am FPGA
     /* 2 = Leiterplatte 2.0 FPGA-CORE=2.5V,Trimmer
       21 = Leiterplatte 2.1 FPGA-CORE=2.5V,Relais (R15,R16 Originalbestueckung des Jahres 2005) */
     if(ResultNINIT[minor]!=0)
	printk(KERN_WARNING "%s hwversion=%d cardyear=%d  error: ResultNINIT=%d\n",XADIO_NAME(minor),xadio_table[minor].eeprom.hwversion,xadio_table[minor].eeprom.cardyear,ResultNINIT[minor]);
     else	
  	xadio_redled(minor,0);
   }  

   xadio_setmodeword(minor,xadio_getmodeword(minor)); /* AIN */
   return 0;
}

/*
static int xadio_eepromexport(int minor) {
  return 0;
}

static int xadio_eepromflush(void) {
  return 0;
}


static int xadio_geteepromversion(int minor,int channel) {
  return xadio_table[minor].eeprom.version;
}

static int xadio_geteepromhardwarerevision(int minor,int channel) {
  return xadio_table[minor].eeprom.hwversion;
}

static int xadio_geteepromcardyear(int minor,int channel) {
  return xadio_table[minor].eeprom.cardyear;
}

static int xadio_geteepromcardnumber(int minor,int channel) {
  return xadio_table[minor].eeprom.cardnumber;
}

static int xadio_geteepromjumpers(int minor,int channel) {
  if(channel<0 || channel>=XADIO_CHANNELS) return 0;
  return xadio_table[minor].eeprom.jumpers[channel];
}
*/

static int xadio_registerfreq(int minor, int freq)
{
int            diff,freqindex;
unsigned int   i,channel;

   freqindex = 0;
   for(i=1; i<xadio_table[minor].eeprom.entries; i++){
      diff = abs(xadio_table[minor].eeprom.entry[i].frequency-freq);
      if(diff<=abs(xadio_table[minor].eeprom.entry[freqindex].frequency-freq))
         freqindex = i;
   }
   xadio_table[minor].eeprom.curr_freq=xadio_table[minor].eeprom.entry[freqindex].frequency;
   if(debug>0) printk(KERN_DEBUG "%s xadio_registerfreq:%d index=%d/%d\n",XADIO_NAME(minor),xadio_table[minor].eeprom.entry[freqindex].frequency,freqindex,xadio_table[minor].eeprom.entries);
   for(channel=0; channel<XADIO_CHANNELS; channel++){
      xadio_table[minor].eeprom.adccorr[channel].dil = xadio_table[minor].eeprom.jumpers[channel];
      xadio_table[minor].eeprom.adccorr[channel].offs0= xadio_table[minor].eeprom.entry[freqindex].corr0.offsets[channel];
      xadio_table[minor].eeprom.adccorr[channel].offs1= xadio_table[minor].eeprom.entry[freqindex].corr1.offsets[channel];
      if(xadio_table[minor].eeprom.jumpers[channel] & 1){ /* bipolar */
         xadio_table[minor].eeprom.adccorr[channel].xm    =0x8000;
         switch(xadio_table[minor].eeprom.jumpers[channel]){
                case 11: 
    case  9: xadio_table[minor].eeprom.adccorr[channel].uend  =20000; break; /* 20000(000) uV = 64(000) digits */
                case  7:
    case  5: xadio_table[minor].eeprom.adccorr[channel].uend  =10000; break;
                case  3:
    case  1: xadio_table[minor].eeprom.adccorr[channel].uend  = 5000; break;
         }
      }
      else{ /* unipolar */
         xadio_table[minor].eeprom.adccorr[channel].xm   =768;
         switch(xadio_table[minor].eeprom.jumpers[channel]){
    case  10: xadio_table[minor].eeprom.adccorr[channel].uend  =20000; break;
    case   6: xadio_table[minor].eeprom.adccorr[channel].uend  =10000; break;
    case   2: xadio_table[minor].eeprom.adccorr[channel].uend  = 5000; break;
         }
      }
      if(debug>0) printk(KERN_DEBUG "%s adio_registerfreq[%d]: offs0:%d offs1:%d xm:%d\n",XADIO_NAME(minor),channel,xadio_table[minor].eeprom.adccorr[channel].offs0,xadio_table[minor].eeprom.adccorr[channel].offs1,xadio_table[minor].eeprom.adccorr[channel].xm);
   }

   return freqindex;
}

static int xadio_adccorr(int minor,int channel,int adc) {
xadio_adccorr_t *curradccorr;
int d0,result;
  curradccorr=&(xadio_table[minor].eeprom.adccorr[channel]);
  if((curradccorr->dil&3)==2) {
    adc=0x10000-adc;
  }
  d0=adc-curradccorr->offs0-curradccorr->xm;
  result=d0+((d0*curradccorr->offs1)/0x8000)+curradccorr->xm;
  if(debug>=4){
    int uv=xadio_value2uv(minor,channel,result);
    printk(KERN_DEBUG "%s xadio_adccorr[%d]:0x%04x->0x%04x %5d->%5d(%d) uv=%d offs0:%d offs1:%d var:%d xm=%d d0=%d\n",XADIO_NAME(minor),channel,adc,result,adc,result,result-adc,uv,curradccorr->offs0,curradccorr->offs1,(d0*curradccorr->offs1)/0x8000,curradccorr->xm,d0);
  }
  return result;
}

int xadio_get_dac_corr(int minor,int dac,int uV)
{
int Result;
int uprec=2;
int PREC_FACT=2;

  if(xadio_table[minor].eeprom.dac_m[dac]==0)
     Result=uV;
  else{
     uV = uV*uprec/1000;
     Result  = uV*xadio_table[minor].eeprom.dac_m[dac]/0x8000 + uprec*xadio_table[minor].eeprom.dac_y0[dac];
     Result  = Result*1000/PREC_FACT/uprec;
  }
  if(debug){
     printk(KERN_DEBUG "%s dac_corr%d %d->%duV\n",XADIO_NAME(minor),dac,uV,Result);
  }
  return Result;
}

/* ------------------------- PLX         ----------------------------- */


#define XADIO_ADDITIONAL_RESET
static void plx_init(int minor) {
#ifdef XADIO_ADDITIONAL_RESET
unsigned char save_irq;
  pci_read_config_byte(xadio_table[minor].pdev,0x3c,&save_irq);
  SetPlx_cntrl(minor,GetPlx_cntrl(minor)|0x40000000); /* Reset on */
  xadio_sleep(minor,20);
  SetPlx_cntrl(minor,GetPlx_cntrl(minor)&~0x40000000); /* Reset off */
  xadio_sleep(minor,20);
  SetPlx_cntrl(minor,GetPlx_cntrl(minor)|0x20000000); /* Load EEPRom on */
  xadio_sleep(minor,20);
  SetPlx_cntrl(minor,GetPlx_cntrl(minor)&~0x20000000); /* Load EEPRom off */
  xadio_sleep(minor,20);
  pci_write_config_byte(xadio_table[minor].pdev,0x3c,save_irq);
#endif
  pci_write_config_word(xadio_table[minor].pdev,0x06,0xfb90);/*reset all errors*/
  pci_write_config_byte(xadio_table[minor].pdev,0x04,0x16); /* Allow write and inval , forbid IO */
  SetPlx_dmacsr0(minor,0x0); /* disable -stop -resirq  */
  SetPlx_dmacsr1(minor,0x0); /* disable -stop -resirq channel 1 */ 
  /* SetPlx_intcsr(minor,0x00040900); enable pciint,dmaint,locintin 0000.0000.0000.0100.0000.1001.0000.0000 */

  /* 0000.0000.0000.0010.1011.0110.0100.0011

  0000000000000 RESERVERD
  CHAIN DISABLED (64BIT) 0
  INTERRUPT TO PCI 1
  CLEAR COUNT DISABLED 0.
  FAST/SLOW TERM FAST 1.
  DMA EOT DISABLED 0.
  MEMORY WRITE AND INVAL ENAB 1.
  DEMAND MODE ENABLED 1.
  LOCALADRESSING INCR 0.
  DONEINT ENABLED 1.
  SCATHER ENABLED 1.
  BURST DISABLED 0.
  BTERM DISABLED 0.
  READY ENABLED 1.
  1WAIT 0001.
  32BIT 11
 */
  SetPlx_dmamode0(minor,0x0002b647); /* 32Bit 1 wait ready scather/gather from pci */
  SetPlx_marbr(minor,0x01280000); 
  SetPlx_lbrd0(minor,0x42430147);        /* _mo_ 20.10.2003 0x42430143->0x42430147 more wait states */
                                        /* to prevent DAC,DIGIO,... registers from being randomly overwritten while dma is running */
}

static void xadio_plxintmode(int minor,int enable) {
   SetPlx_intcsr(minor,enable?0x00040900:0); /* enable pciint,dmaint,locintin 0000.0000.0000.0100.0000.1001.0000.0000 */
}

static void xadio_dmastart(int minor) {
  if(debug) printk(KERN_DEBUG "%s xadio_dmastart\n",XADIO_NAME(minor));
  /* 0000.0000.0000.0011.1011.0110.0100.0011*/
  /*  0000.0000.0000.0020.10 WINV 1.0 IEN S/G 0.0100.0011 */
  SetPlx_dmapadr0(minor,0);
  SetPlx_dmaladr0(minor,0);
  SetPlx_dmasiz0(minor,0);
  SetPlx_dmadpr0(minor,xadio_table[minor].darr[bcount-1]->next);
  SetPlx_dmacsr0(minor,0x3);
  xadio_table[minor].flags|=XADIO_DMA_MASK;
}

static void xadio_dmastop(int minor) {
  if(debug) printk(KERN_DEBUG "%s xadio_dmastop\n",XADIO_NAME(minor));
  SetPlx_dmacsr0(minor,0xc);
  xadio_table[minor].flags&=~(XADIO_DMA_MASK);
  xadio_sleep(minor,50);
}

static int xadio_getvarcore(char *buffer,int minor,int offset,int type,int size,int space) {
int len=0;
  switch(type) {
    case 0: {
      if(size>4096) size=4096;
      switch(space) {
        case 0: 
          memcpy_fromio(buffer,xadio_table[minor].base0+offset,len=size);
          break;
        case 1: 
          memcpy_fromio(buffer,xadio_table[minor].base2+offset,len=size);
          break;
        case 2:
          memcpy(buffer,((char*)(&xadio_table[minor]))+offset,len=size);
          break;
        case 3:
          break;
      }
      }
      break;
    case 1: {
      unsigned char val=0xba;
      switch(space) {
        case 0: 
          val=readb(xadio_table[minor].base0+offset);
          break;
        case 1: 
          val=readb(xadio_table[minor].base2+offset);
          break;
        case 2:
          val=*((u8*)(((char*)(&xadio_table[minor]))+offset));
          break;
        case 3:
          pci_read_config_byte(xadio_table[minor].pdev,offset,&val);
          break;
      }
      len=sprintf(buffer,"0x%2.2x",val);
      }
      break;
    case 2:{
      unsigned short val=0xbadb;
      switch(space) {
        case 0: 
          val=readw(xadio_table[minor].base0+offset);
          break;
        case 1: 
          val=readw(xadio_table[minor].base2+offset);
          break;
        case 2:
          val=*((u16*)(((char*)(&xadio_table[minor]))+offset));
          break;
        case 3:
          pci_read_config_word(xadio_table[minor].pdev,offset,&val);
          break;
      }
      len=sprintf(buffer,"0x%4.4x",val);
      }
      break;
    case 4: {
      unsigned int val=0xbadbadba;
      switch(space) {
        case 0: 
          val=readl(xadio_table[minor].base0+offset);
          break;
        case 1: 
          val=readl(xadio_table[minor].base2+offset);
          break;
        case 2:
          val=*((u32*)(((char*)(&xadio_table[minor]))+offset));
          break;
        case 3:
          pci_read_config_dword(xadio_table[minor].pdev,offset,&val);
          break;
      }
      len=sprintf(buffer,"0x%8.8x",val);
      }
      break;
  }
  return len;
}

static int xadio_scanint(const char *buffer,int len,int *result) {
char lbuffer[20];
char *endptr;
int sign=1;
  if(len>0 && buffer[0]=='-') {
    sign=-1;
    buffer++;
    len--;
  }
  len=MIN(len,(int)sizeof(lbuffer)-1);
  memcpy(lbuffer,buffer,len);
  lbuffer[len]=0;
  *result=simple_strtoul(lbuffer,&endptr,0)*sign;
  if(!(*endptr) || isspace(*endptr)) return 0;
  return -ERANGE;
}

static int xadio_scanhex(const char *buffer,int len,char *destination,int size,int *lenPtr) {
int i,currVal=0,pos=0;
int position=0;
  *lenPtr=0;
  for(i=0;i<len;i++) {
    int val=buffer[i];
    if(isspace(val)) continue;
    switch(val) {
      case '0'...'9': val-='0'; break;
      case 'A'...'F': val-='A'-10; break;
      case 'a'...'f': val-='a'-10; break;
      default:
        return -EINVAL;
    }
    if(position&1) {
      currVal=(currVal<<4)|(val&0xf);
      if(pos<size) {
        destination[pos++]=currVal;
      }
    } else {
      currVal=val&0xf;
    }
    position++;
  }
  if(position&1) {
    return -EINVAL;
  }
  *lenPtr=pos;
  return 0;
}

static int xadio_setvarcore(const char *buffer,int count,int minor,int offset,int type,int size,int space) {
int value=0;
  switch(type) {
    case 1:
    case 2:
    case 4:
      if(xadio_scanint(buffer,count,&value)) return -ERANGE;
  }
  switch(space) {
    case 0:
      switch(type) {
         case 0:
           if(count>size) count=size;
           memcpy_toio(xadio_table[minor].base0+offset,buffer,count);
           break;
         case 1:
           writeb(value,xadio_table[minor].base0+offset);
           break;
         case 2:
           writew(value,xadio_table[minor].base0+offset);
           break;
         case 4:
           writel(value,xadio_table[minor].base0+offset);
           break;
      }
      break;
    case 1:
      switch(type) {
         case 0:
           if(count>size) count=size;
           memcpy_toio(xadio_table[minor].base2+offset,buffer,count);
           break;
         case 1:
           writeb(value,xadio_table[minor].base2+offset);
           break;
         case 2:
           writew(value,xadio_table[minor].base2+offset);
           break;
         case 4:
           writel(value,xadio_table[minor].base2+offset);
           break;
      }
      break;
    case 2:
      switch(type) {
         case 0:
           if(count>size) count=size;
           memcpy(((char*)(&xadio_table[minor]))+offset,buffer,count);
           break;
         case 1:
           *(((u8*)(&xadio_table[minor]))+offset)=value;
           break;
         case 2:
           *((u16*)(((char*)(&xadio_table[minor]))+offset))=value;
           break;
         case 4:
           *((u32*)(((char*)(&xadio_table[minor]))+offset))=value;
           break;
      }
      break;
    case 3:
      switch(type) {
        case 1:
          pci_write_config_byte(xadio_table[minor].pdev,offset,value);
          break;
        case 2:
          pci_write_config_word(xadio_table[minor].pdev,offset,value);
          break;
        case 4:
          pci_write_config_dword(xadio_table[minor].pdev,offset,value);
          break;
      }
    break;
  }
  return count;
}

/* ---------------------- CPLD-Stuff --------------------------------- */

/*----------------------------------------------
 JTAG Programmierung des XILINX CPLD 
Pin     anzuschliessendes Pin
1        GND
2        TDO
3        TDI
4        TCK
5        TMS
6        VDD (3.3V)
-----------------------------------------------*/

/* CPLD for FPGA - Programming:
 Adresse:
 0x80000000 - 10xxxxxx xxxxxxxx xxxxxxxx xxxxx0BB:
  Register 0:
    CPLD_DATA
      D7 .... D0 Gespiegelt zum Bit-File !! 
  Register 1:
 Adresse:
 0x80000004 - 10xxxxxx xxxxxxxx xxxxxxxx xxxxx1BB:
  Register 4:
   CPLD_CONTROL:
 
   BIT 0 - NPROG  (RW)
   BIT 1 - NCCS   (RW)
   BIT 2 - NINIT  (R)
   BIT 3 - DONE   (R)
   BIT 4 - CCLK   (RW)
   BIT 5 - NCWR   (RW)
   BIT 6 - NLED   (RW)
   BIT 7 - USER0  (RW)
*/

  
#define SetCPLD_Data(minor,value)     WFL(minor,0,((value)&0xff))
#define SetCPLD_Control(minor,value)  WFL(minor,4,((value)&0xff))
#define GetCPLD_Data(minor)           (RFL(minor,0)&0xff)
#define GetCPLD_Control(minor)        (RFL(minor,4)&0xff)

#define XADIO_CPLD_NPROG_BIT           0x01
#define XADIO_CPLD_NCCS_BIT            0x02
#define XADIO_CPLD_NINIT_BIT           0x04
#define XADIO_CPLD_DONE_BIT            0x08
#define XADIO_CPLD_CCLK_BIT            0x10
#define XADIO_CPLD_NCWR_BIT            0x20
#define XADIO_CPLD_NLED_BIT            0x40
#define XADIO_CPLD_USER_BIT            0x80

#define XADIO_CPLD_BASEADDR                0x80000000

/* master/single */
/* static unsigned char xadio_fpga_data[] __initdata = */
static unsigned char xadio_fpga_data[] =
#include "xadio_fpga.h"
;
/* slave */
#if (MINOR_MASTER>=0)
 static unsigned char xadio_fpga_datas[] =
 #include "xadio_fpgas.h"
 ;
#endif

static inline void xadio_delayCF(int minor,int delay) {
  wmb();
  if(delay) udelay(delay);
}

static inline int xadio_setCF(int minor,u8 value) {
u8 control;
  if(debug>2) printk(KERN_DEBUG "%s xadio_setCF:0x%2.2x\n",XADIO_NAME(minor),value);
  SetCPLD_Control(minor,value);
  xadio_delayCF(minor,1);
  if(((control=GetCPLD_Control(minor))&0xf3)==(value&0xf3)) return 0;
  printk(KERN_WARNING "%s Error in xadio_setCF:0x%2.2x!=0x%2.2x\n",XADIO_NAME(minor),control,value);
  return -EIO;
}

static inline int xadio_getCF(int minor) {
int control;
  control=GetCPLD_Control(minor);
  if(debug>2) printk(KERN_DEBUG "%s xadio_getCF:0x%2.2x\n",XADIO_NAME(minor),control);
  return control;
}

static inline int xadio_checkCFMasked(int minor,u8 value,u8 mask) {
u8 control=xadio_getCF(minor);
  if((control&mask)==(value&mask)) return 0;
  printk(KERN_WARNING "%s xadio_checkCF:0x%2.2x!=0x%2.2x\n",XADIO_NAME(minor),control,value);
  return -EIO;
}

static inline int xadio_checkCF(int minor,u8 value) {
  return xadio_checkCFMasked(minor,value,0xff);
}

static inline int xadio_writeCF(int minor,u8 value) {
  if(debug>2) printk(KERN_DEBUG "%s writeCF:0x%2.2x\n",XADIO_NAME(minor),value);
  SetCPLD_Data(minor,value); /* Write Data Register */
  xadio_delayCF(minor,0);
  if(GetCPLD_Data(minor)==value) return 0;
  printk(KERN_WARNING "%s Error in writeCF:0x%2.2x!=0x%2.2x\n",XADIO_NAME(minor),value,GetCPLD_Data(minor));
  return -EIO;
}

static int xadio_progfpga(int minor) {
unsigned int save_base0;
int i;
int result=-EIO;

  xadio_redled(minor,1);
  if(hwversion<2) return 0;
  if(debug) printk(KERN_DEBUG "%s xadio_progfpga:\n",XADIO_NAME(minor));
  save_base0=GetPlx_las0ba(minor);                   /* Save Pointer to FPGA */
  SetPlx_las0ba(minor,XADIO_CPLD_BASEADDR|0x1);          /* Set Pointer to Config_FPGA + VALID */
  xadio_delay();
  if(xadio_setCF(minor,0x63)) goto error;    /* NPROG high NCCS high CCLK low LED high NCWP high */
  if(xadio_setCF(minor,0x62)) goto error;    /* NPROG low  High:NLED,NCWR,NCCS  FPGA:DONE->0 NINIT->0 */
  if(xadio_checkCF(minor,0x62)) goto error;
  if(xadio_setCF(minor,0x63)) goto error;    /* NPROG high */
  xadio_delayCF(minor,5);                    /* Wait 5 us for Initialization */
  if(xadio_checkCFMasked(minor,0x67,0xfb)) goto error;  /* NINIT high  High:NLED,NCWR,NCCS FPGA:NINIT->1 alles ausser ninit testen?! NPROG==1,NCCS==1,DONE==1*/
  if(xadio_setCF(minor,0x41)) goto error;    /* NCWR low NCCS low */
  if(xadio_checkCF(minor,0x45)) goto error; 
  for(i=0;i<ARRSIZE(xadio_fpga_data);i++) {
    int value;
    #if (MINOR_MASTER>=0)	
    if(minor!=MINOR_MASTER)	
      value=xadio_fpga_datas[i];
    else
    #endif	
      value=xadio_fpga_data[i];
    if(xadio_writeCF(minor,value)) goto error; /* Write Data Register */
    if(xadio_setCF(minor,0x51)) goto error;    /* CLOCK HIGH */
    if(xadio_setCF(minor,0x41)) goto error;    /* CLOCK LOW */
    if(xadio_getCF(minor)&0x8) break;          /* Done ? */
    if(!(i%10240)) xadio_sleep(minor,10);      /* schedule sometimes 10 */
  }
  xadio_sleep(minor,20);
  if(xadio_checkCF(minor,0x4d)) goto error;    /* DONE NINIT */
  for(i=0;i<8;i++) {                           /* Extra Clocks for Phase 1-4 */
    if(debug>2) printk(KERN_DEBUG "%s xadio_progfpga: extra clocks:%d\n",XADIO_NAME(minor),i);
    if(xadio_setCF(minor,0x51)) goto error;    /*  CLOCK HIGH */
    if(xadio_setCF(minor,0x41)) goto error;    /* CLOCK LOW */
  }
  if(xadio_setCF(minor,0x63)) goto error;      /* Setze auf NOT_PROG/NOT_CCS/LED off */
  /* Die folgende Zeile war falsch und geht nicht mehr mit der korrigierten FPGA-Core Spannung 1.8V */  
  /*if(xadio_checkCF(minor,0x6b)) goto error;*/    /* DONE not NINIT */
  if(xadio_checkCFMasked(minor,XADIO_CPLD_DONE_BIT,XADIO_CPLD_DONE_BIT))   goto error;   /* XADIO_CPLD_DONE_BIT==1 */
  ResultNINIT[minor]=xadio_getCF(minor) & XADIO_CPLD_NINIT_BIT;   /* spaeter testen, wenn Kartennummer bekannt ist */

  result=0;
error:
  xadio_setCF(minor,0x63);                     /* Setze auf NOT_PROG/NOT_CCS/LED off */
  SetPlx_las0ba(minor,save_base0);             /* Restore Pointer to FPGA */
  xadio_delay();
  if(debug) printk(KERN_DEBUG "%s xadio_progfpga: result:%d\n",XADIO_NAME(minor),result);

  #if (MINOR_MASTER>=0)
  if(minor==MINOR_MASTER)
  	xadio_master(minor); /*not ready*/
  #endif
  xadio_table[minor].glFpgaLoaded=result==0?1:0;
  return result;
}

static int xadio_cpldled(int minor,int onoff) {
unsigned int save_base0;
u8 control;
  if(nosubvendorcheck) return 0;
  if(hwversion<2) return 0;
  if(debug) printk(KERN_DEBUG "%s xadio_cpldled: %d\n",XADIO_NAME(minor),onoff);
  save_base0=GetPlx_las0ba(minor);               /* Save Pointer to FPGA */
  SetPlx_las0ba(minor,XADIO_CPLD_BASEADDR|0x1);  /* Set Base0 to Config_FPGA + VALID */
  xadio_delay();
  control=xadio_getCF(minor);
  control&=~(XADIO_CPLD_NLED_BIT);               /* LED on */
  control|=onoff?0:XADIO_CPLD_NLED_BIT;          /* LED off if not onoff */
  xadio_setCF(minor,control);
  SetPlx_las0ba(minor,save_base0);               /* Restore Base0 pointer */
  xadio_delay();
  return 0;
}

/* -------------------- FPGA-Register -------------------------------- */

/*
FPGA Control Register
  +---------------------------------------------------------------------+
  |REC |REI |REA |CAL |X   |RT6 |RT5|RT4|RT3|RT2|RT1|RT0|ID0|IP0|ID1|IP1|
  +---------------------------------------------------------------------+
  |IO15|IO14|IO13|IO12|IO11|IO10|IO9|IO8|IO7|IO6|IO5|IO4|IO3|IO2|IO1|IO0|
  +---------------------------------------------------------------------+
     REC=Resetcounter
     REI=ResetInterrupt
     REA=ResetADC
     CAL=Calibrate
     RT0..6=Rate (Teiler)
     ID0=Icrementalgeber (Double Mode) Kanal 0 (0 normal)
     IP0=Icrementalgeber (Polaritaet) Kanal 0 (0 high aktiv)
     ID1=Icrementalgeber (Double Mode) Kanal 1
     IP1=Icrementalgeber (Polaritaet) Kanal 1
     IO
*/

static unsigned int xadio_getmodeword(int minor) {
unsigned int modeword;
  modeword=GetFpga_cntrl(minor);
  if(debug>3) printk(KERN_DEBUG "%s xadio_getmodeword::%8.8x\n",XADIO_NAME(minor),modeword);
  return modeword;
}

static void xadio_setmodeword(int minor,unsigned int modeword) {
int use_analog_inverting;  	/* nicht mehr 1/0 sonden 1/20 */
static int use_analog_inverting_msg[XADIO_NO];		 
 
  modeword &= ~0x200; /* default is xadio >= Rev. 2.1 (no inverting) */
  if(xadio_table[minor].eeprom.cardyear>0){
    use_analog_inverting=xadio_table[minor].eeprom.cardyear>=5?0:1;
    if(use_analog_inverting)
      modeword |=  0x200;
    else
      modeword &= ~0x200;
  }
  else
    use_analog_inverting=(modeword&0x200)==0?0:1; /* if(debug>0)printk(KERN_DEBUG "%s cardyear=0 -> using default AIN!\n",XADIO_NAME(minor)); */
  if(use_analog_inverting_msg[minor]<10){
    use_analog_inverting_msg[minor]++;	
    printk(KERN_DEBUG "%s cardyear=%d -> use_analog_inverting=%d (AIN).\n",XADIO_NAME(minor),xadio_table[minor].eeprom.cardyear,use_analog_inverting);
  }	
  if(debug>4) printk(KERN_DEBUG "%s xadio_setmodeword::%8.8x\n",XADIO_NAME(minor),modeword);
  SetFpga_cntrl(minor,modeword);
  xadio_delay();
}

static void fpga_init(int minor) {
  #if (MINOR_MASTER>=0)
  if(minor==MINOR_MASTER) /* not ready */
	  xadio_setmodeword(minor,0x0800f00f);  /* Current Initial Hardware Config */
  else
  #endif
	  xadio_setmodeword(minor,0xf00f);  /* Current Initial Hardware Config */
  //SetFpga_ien(minor,0x00100000);  /* Enable Incr-Null-Interrupt */
  SetFpga_div0(minor,0);
  SetFpga_div1(minor,0);
  /*SetFpga_cntrl(minor,GetFpga_cntrl(minor)|0x20000);*/
}

#if (MINOR_MASTER>=0)
static void xadio_master(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,0x08000000 | modeword); /* Bit 27, enable output clock  */
}
#endif

static void xadio_calibrateon(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,0x10000000 | modeword); /* Calib ADC */
}

static void xadio_calibrateoff(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,0xefffffff & modeword); /* Calib ADC off */
}

static void xadio_redled(int minor,int onoff) {
unsigned int mode;
  if(!xadio_table[minor].glFpgaLoaded)
    printk(KERN_DEBUG "%s xadio_redled failed\n",XADIO_NAME(minor));
  else{  
      printk(KERN_DEBUG "%s xadio_redled ok\n",XADIO_NAME(minor));
      mode=xadio_getmodeword(minor);  
      mode=(mode&(~(0x20)))|(onoff?0x20:0);
      xadio_setmodeword(minor,mode);
  }  
}

static void xadio_greenled(int minor,int onoff) {
unsigned int mode;
  if(!xadio_table[minor].glFpgaLoaded)
    printk(KERN_DEBUG "%s xadio_greenled failed\n",XADIO_NAME(minor));
  else{  
    printk(KERN_DEBUG "%s xadio_greenled ok\n",XADIO_NAME(minor));
    mode=xadio_getmodeword(minor);  
    mode=(mode&(~(0x40)))|(onoff?0x40:0);
    xadio_setmodeword(minor,mode);
  }
}

static void xadio_calibrate(int minor) {
  if(debug) printk(KERN_DEBUG "%s xadio_calibrate\n",XADIO_NAME(minor));
  xadio_calibrateon(minor);
  xadio_calibrateoff(minor);
}

static void xadio_resetaduon(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,0x20000000 | modeword); /* reset ADCs */
}

static void xadio_resetaduoff(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,modeword &0xdfffffff); /* reset ADCs */
}

static void xadio_resetadu(int minor) {
  xadio_resetaduon(minor);
  xadio_resetaduoff(minor);
}

static void xadio_resetint(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  xadio_setmodeword(minor,0x40000000 | modeword); /* reset INT */
  xadio_setmodeword(minor,0xbfffffff & modeword); /* reset INT off*/
}

static void xadio_resetcounteron(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  modeword |= 0x80000000;           /* reset counter on */
  xadio_setmodeword(minor,modeword);
}

static void xadio_resetcounteroff(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
  modeword &= 0x7fffffff;           /* reset counter off */
  xadio_setmodeword(minor,modeword);
}

static void xadio_resetcounter(int minor) {
  xadio_resetcounteron(minor);
  xadio_resetcounteroff(minor);
  xadio_table[minor].counter0=-1;
  xadio_table[minor].counter1=0;
}

static void xadio_setrate(int minor,int rate) {
unsigned int modeword=xadio_getmodeword(minor);
int rate0=rate;  
  if(xadio_getfpgaversion(minor,0)>=(int)0xadda0210){ /* Vorher war hier ein Fehler im FPGA */
    rate--; 
    if(rate<0) rate=0;
  }
  else
    if(hwversion>1 && rate<1) rate=1;
  modeword=(modeword&0xf80fffff)+((rate&0x7f)<<20);
  if(debug) printk(KERN_DEBUG "%s xadio_setrate %d (corr:%d)\n",XADIO_NAME(minor),rate0,rate);
  xadio_setmodeword(minor,modeword);
}

static int xadio_getrate(int minor) {
unsigned int modeword=xadio_getmodeword(minor);
unsigned int rate,rate0;
  rate0 = rate = (modeword>>20)&0x7f;
  if(xadio_getfpgaversion(0,0)>=(int)0xadda0210)rate++; /* Vorher war hier ein Fehler im FPGA */
  if(debug) printk(KERN_DEBUG "%s xadio_getrate %d (corr:%d)\n",XADIO_NAME(minor),rate0,rate);
  return rate;
}

static const int xadio_quarz=50000000;

static int xadio_base(void) {
  if(hwversion==1) return xadio_quarz/2/128;
  return xadio_quarz/128;
}

static int xadio_rate2freq(int rate) {
  return xadio_base()/(rate+1);
}

static int xadio_freq2rate(int frequency) {
  return (xadio_base()/frequency)-1;
}

static int xadio_getfreq(int minor,int channel) {
  return xadio_rate2freq(xadio_getrate(minor));
}

static int xadio_setfreq(int minor,int frequency) {
int rate;
  if(frequency<3051) frequency=3051;
  if(frequency>195312) frequency=195312;
  rate=xadio_freq2rate(frequency);
  xadio_setrate(minor,rate);
  xadio_registerfreq(minor,xadio_rate2freq(rate)); /* register the current frequency for later digit correction */
#ifdef EXTRA_RESET
  xadio_sleep(minor,2000); 
  xadio_resetadu(minor); /* Makes problems ! */
#endif
/*
  xadio_sleep(minor,200);
  xadio_calibrate(minor);
*/
  xadio_sleep(minor,200);
  return 0;
}

static void xadio_setrpmencoder(int minor,int mode) {
#ifdef USE_RPM_CODE
unsigned int modeword=xadio_getmodeword(minor);
  if(mode==0)	modeword=(modeword & ~0x20000);
  else          modeword=(modeword |  0x20000);	
  xadio_setmodeword(minor,modeword);
#endif
}


static void xadio_setincmode0(int minor,int mode) {
unsigned int modeword=xadio_getmodeword(minor);
#ifdef USE_RPM_CODE
  modeword=(modeword&0xfffeffff)+((mode&1)<<16);	/* bit 17 wird ab Dez06 fuer RPM Umschaltung verwendet */
#else
  modeword=(modeword&0xfffcffff)+((mode&3)<<16);
#endif
  xadio_setmodeword(minor,modeword);
}

static void xadio_setincmode1(int minor,int mode) {
unsigned int modeword=xadio_getmodeword(minor);
  modeword=(modeword&0xfff3ffff)+((mode&3)<<18);
  xadio_setmodeword(minor,modeword);
}

static int xadio_getincmode(int minor,int channel) {
unsigned int modeword=xadio_getmodeword(minor);
  return (modeword>>(channel?18:16))&0x3;
}

static void xadio_setdiv0(int minor,int value) {
  if(debug>0)printk(KERN_DEBUG "%s xadio_setdiv0::%8d -> %8d \n",XADIO_NAME(minor),GetFpga_div0(minor),value);
  SetFpga_div0(minor,value);
}

static void xadio_setdiv1(int minor,int value) {
  if(debug>0)printk(KERN_DEBUG "%s xadio_setdiv1::%8d -> %8d \n",XADIO_NAME(minor),GetFpga_div1(minor),value);
  SetFpga_div1(minor,value);
}

static int xadio_getdiv(int minor,int channel) {
  if(channel==0) return GetFpga_div0(minor);
  else           return GetFpga_div1(minor);
}


static void xadio_setdigio(int minor,u32 value,u32 mask) { 
u32 orig=GetFpga_digio(minor) & ~mask;
  if(debug>1) printk(KERN_DEBUG "%s xadio_setdigio Value: %x Mask: %x\n",XADIO_NAME(minor),value,mask);
  SetFpga_digio(minor,orig | (value&mask));
}

static void xadio_settestmode(int minor,int mode) {
unsigned int modeword=xadio_getmodeword(minor);
  modeword=(modeword&0xfffffe7f)+((mode&3)<<7);
  xadio_setmodeword(minor,modeword);
}


static int xadio_reset(int minor) {
int result=0;
  xadio_dmastop(minor); /* mo 20.12.2004 line added,  otherwise suse 92 hangs after reset */
  plx_init(minor);
  if(!nosubvendorcheck){
    if((result=xadio_progfpga(minor))==0){
      fpga_init(minor);
      xadio_resetcounter(minor);
      xadio_sleep(minor,20);
      xadio_setfreq(minor,195312);
      xadio_plxintmode(minor,1);
      xadio_dmastart(minor);
      xadio_setdacuv(minor,0,0);
    }
  }
  return result;
}

static inline unsigned int xadio_countersample(int minor,int bnr,int offset) {
  return ((Xadio_record*)xadio_roundptr16(xadio_table[minor].barr[bnr]))[offset].timer32;
}

static inline int xadio_oktant(unsigned int counter) {
  return (counter&0x70000000)>>28;
}

static inline void xadio_corrcounter(int minor) {
  unsigned int counter=xadio_countersample(minor,0,0);
  switch(xadio_oktant(counter)) {
    case 0:
      if(xadio_table[minor].counter1==xadio_table[minor].counter0) {
        xadio_table[minor].counter1++;
      }
      break;
    case 4:
      if(xadio_table[minor].counter1!=xadio_table[minor].counter0) {
        xadio_table[minor].counter0=xadio_table[minor].counter1;
      }
      break;
  }
}

static inline void xadio_aktbuff(int minor) {                 
unsigned int dst=GetPlx_dmapadr0(minor);  /*  wird jetzt oder in Zukunft beschrieben */                    
int j;                                                        
  for(j=0;j<bcount;j++) {                                     
    if(dst==xadio_table[minor].darr[j]->dst) {                
      xadio_table[minor].aktbuf=j;                            
      return;                                                 
    }                                                         
  }                                                           
}                                                             
                                                              
#ifdef USE_DEBUG_IN_CONTROL
static void xadio_check_ring(int minor)
{
int timer32,last32,b,diff,ok;
char tmp[256];

if(xadio_table[minor].debug_interrupts>1000){
   last32=xadio_countersample(minor,bcount-1,0);
   ok=1; 
   tmp[0]=0;
   sprintf(tmp,"%d, ",xadio_table[minor].aktbuf);
   for(b=0;b<bcount;b++) {                                     
     timer32=xadio_countersample(minor,b,0);
     sprintf(tmp+strlen(tmp),"%d:%d ",timer32,timer32-last32);
     diff=(timer32-last32)<0 ? (last32-timer32) : (timer32-last32);
     if(diff%1024!=0)
	ok=0;
     last32=timer32;
   }
     if(!ok)
	strcpy(xadio_table[minor].debugtext,tmp);	
}
}
#endif

static void xadio_localint(int minor,int irqsource) {
  if (waitqueue_active(&(xadio_table[minor].local_wait))) wake_up_interruptible(&(xadio_table[minor].local_wait));
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static irqreturn_t xadio_interrupt(int irq, void *dev_id , struct pt_regs *regs) { /* DMA fertig mit 1 Block oder andere events */
#else
static irqreturn_t xadio_interrupt(int irq, void *dev_id) { /* DMA fertig mit 1 Block oder andere events */
#endif
unsigned int minor=ARRPOS((xadio_t*)dev_id,xadio_table);
unsigned int intcsr;
int handled=0;

#ifdef USE_DEBUG_IN_CONTROL
  xadio_table[minor].debug_interrupts++;
#endif
  intcsr=GetPlx_intcsr(minor);
  if(intcsr & 0x8000) { /* local interrupt active */ 
    unsigned int irqsource=GetFpga_isrc(minor); /* read clears irq ? */
    if(irqsource&0x00100000) {
    }
    xadio_localint(minor,irqsource);
    handled++;
  }
  if(intcsr & 0x200000) { /* dma0 interrupt active , neuer Block */
    xadio_corrcounter(minor);
    xadio_aktbuff(minor);                                     
#ifdef USE_DEBUG_IN_CONTROL
    xadio_check_ring(minor);
#endif
#ifndef NO_RESTART_DMA
    if((xadio_table[minor].flags)&XADIO_DMA_MASK) {
      SetPlx_dmacsr0(minor,0x9); /* resirq + enable */
    } else {
      SetPlx_dmacsr0(minor,0x8); /* resirq */
    }
#endif
    if (waitqueue_active(&(xadio_table[minor].dma_wait))) wake_up_interruptible(&(xadio_table[minor].dma_wait));
    handled++;
  }
  return IRQ_RETVAL(handled);
}

/* Device-IO Functions */

static ssize_t xadio_write(struct file * file, const char * buf, size_t count, loff_t *ppos) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
  if(debug) printk(KERN_DEBUG "%s write: count:%d\n",XADIO_NAME(minor),(int)count);
  return count;
}

#ifdef EMU64
static inline loff_t xadio_corrpos(int minor,unsigned int counter) {
loff_t offset=0;
  switch(xadio_oktant(counter)) {
    case 0:
    case 1:
      offset=xadio_table[minor].counter0+1;
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      offset=xadio_table[minor].counter1;
      break;
    case 6:
    case 7:
      offset=xadio_table[minor].counter0;
      break;
  }
  return ((offset<<31)|(counter&0x7fffffff))<<6;
}
#else

static inline loff_t xadio_corrpos(int minor,unsigned int counter) {
loff_t offset=GetFpga_timer64(minor);
  return ((offset<<32)|(counter&0xffffffff))<<6;
}
#endif


static inline loff_t xadio_getpos(int minor,int bnr) {
  return xadio_corrpos(minor,xadio_countersample(minor,bnr,0));
}

static inline loff_t xadio_getlst(int minor,int bnr) {
  return xadio_corrpos(minor,xadio_countersample(minor,bnr,brecords-1));
}

static void xadio_debugbuffers(int minor) {
int i;
  for(i=0;i<bcount;i++) {
    loff_t x1=xadio_getpos(minor,i);
    loff_t x2=xadio_getlst(minor,i);
    printk(KERN_WARNING "%s buffer[%d] 0x%8.8x%8.8x 0x%8.8x%8.8x\n",XADIO_NAME(minor),i,HILO(x1),HILO(x2));
  }
}

static int xadio_findbuffer(int minor,loff_t pos,loff_t *min,loff_t *max) {
int result=-1;
int i;
  *min=0x7fffffffffffffffLL;
  *max=-1LL;
  for(i=0;i<bcount;i++) {
    loff_t p1=xadio_getpos(minor,i);
    loff_t p2=xadio_getlst(minor,i);
    if(pos>=p1 && pos<=p2) result=i;
    if(p1<*min) { *min=p1; } 
    if(p2>*max) { *max=p2; } 
  }
  return result;
}

static int xadio_checkbuffer(int minor,int bnum,loff_t pos) {
loff_t p1=xadio_getpos(minor,bnum);
loff_t p2=xadio_getlst(minor,bnum);
  if(pos>=p1 && pos<=p2) return 1;
  return 0;
}

static inline int xadio_findlowbuffer(int minor) {
int i;
int result=0;
loff_t min=0x7fffffffffffffffLL;
  for(i=0;i<bcount;i++) {
    loff_t p1=xadio_getpos(minor,i);
    if(p1<min) { min=p1; result=i; } 
  }
  return result;
}

static inline int xadio_readok(int minor,loff_t pos) {
/*int bnum=xadio_findlowbuffer(minor);*/
int bnum=(((xadio_table[minor].aktbuf)+1)%bcount);            /* int bnum=xadio_findlowbuffer(minor); */
loff_t minpos=xadio_getpos(minor,bnum);
loff_t maxpos=xadio_getlst(minor,(bnum+(bcount-2))%bcount);
  if(pos<minpos) return -1; // overrun
  if(pos>=maxpos+bcount*bsize) return -2; // counterreset
  if(pos>maxpos) return 1;
  return 0;
}

static ssize_t xadio_read(struct file * file, char * buf, size_t count, loff_t *ppos) { /* read samples */
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
loff_t pos=*ppos;
size_t done=0;
  if(debug>1) printk(KERN_DEBUG "%s read: off:0x%8.8x%8.8x count:%d\n",XADIO_NAME(minor),HILO(pos),count); 
  while(done<count) {
    loff_t oldpos=pos;
    int bnum=(((xadio_table[minor].aktbuf)+1)%bcount);  /* aeltester bufferindex */
    /*int bnum=xadio_findlowbuffer(minor);*/
    loff_t minpos=xadio_getpos(minor,bnum);
    loff_t maxpos=xadio_getlst(minor,(bnum+(bcount-2))%bcount); /* aktueller bufferindex */
    //xadio_corrcounter(minor);
    if(pos<minpos) { /* overrun */
      if(done) break;
      if(debug>1) xadio_debugbuffers(minor);
      pos=minpos;
      if(oldpos) {
        if(debug>0) printk(KERN_WARNING "%s overrun,skipping:0x%8.8x%8.8x starting 0x%8.8x%8.8x maxpos 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(pos-oldpos),HILO(pos),HILO(maxpos));
      }
      continue;
    }
    if(pos>=maxpos+(bsize*bcount)) {  /* strange hop */
      if(debug>1) xadio_debugbuffers(minor);
      pos=maxpos;/* sollte maxpos+1 sein 23.05.07 dann geht seek(now) aber nicht mehr !*/
      if(oldpos) {
        if(debug>0) printk(KERN_WARNING "%s strange underrun,skipping back:0x%8.8x%8.8x starting 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(pos-oldpos),HILO(pos));
      }
      continue;
    }
    if(pos>maxpos) {  /* sleep */
      if(done) break;
      if(file->f_flags & O_NONBLOCK) {
        done=-EAGAIN;
        break;
      }
      interruptible_sleep_on_timeout(&(xadio_table[minor].dma_wait),polling_interval);
      if(signal_pending(current)) {
        done=-EINTR;
        break;
      }
      continue;
    }
    bnum=xadio_findbuffer(minor,pos,&minpos,&maxpos);
    if(bnum>=0) {
      loff_t startpos=xadio_getpos(minor,bnum);
      loff_t offset=pos-startpos;
      size_t size=bsize-offset;
      if(count-done<size) size=count-done;
      copy_to_user(buf+done,xadio_table[minor].barr[bnum]+offset,size);
      if(xadio_checkbuffer(minor,bnum,pos)) {
        pos+=size;
        done+=size;
        break;
      }
      if(done) break;
      pos+=bsize;
      if(debug>0) printk(KERN_WARNING "%s overrun in copy,skipping 0x%8.8x%8.8x starting 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(pos-oldpos),HILO(pos));
      continue;
    }
    if(done) break;
    if(debug>1) xadio_debugbuffers(minor);
    pos+=bsize;
    if(debug>0) printk(KERN_WARNING "%s overrun in search,skipping 0x%8.8x%8.8x starting 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(pos-oldpos),HILO(pos));
    continue;
  }
  *ppos=pos;
  if(debug>1) printk(KERN_DEBUG "%s read: result:%d\n",XADIO_NAME(minor),(int)done);
  return done;
}

static loff_t xadio_nowpos(int minor) {
  return xadio_corrpos(minor,GetFpga_timer(minor));
}

static loff_t xadio_lseek(struct file * file, loff_t offset, int orig) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
loff_t result;
  if(debug>1) printk(KERN_DEBUG "%s lseek: %d 0x%8.8x%8.8x\n",XADIO_NAME(minor),orig,HILO(offset));
  switch (orig) {
    case SEEK_SET:
      result=(file->f_pos=offset);
      break;
    case SEEK_CUR:
      result=(file->f_pos+=offset);
      break;
    case SEEK_END:
      result=(file->f_pos=xadio_nowpos(minor)+offset); /* exact now */
      break;
    default: 
      result=-EINVAL;
      break;
  }
  if(debug>1) printk(KERN_DEBUG "%s lseek(result): 0x%8.8x%8.8x\n",XADIO_NAME(minor),HILO(result));
  return result;
}

static int xadio_open(struct inode * inode, struct file * file) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int minor=MINOR(inode->i_rdev);
#else
struct proc_dir_entry *entry=PDE(inode);
int minor=(int)(entry->data);
  inode->i_rdev=MKDEV(major,minor);
#endif
  if(!XADIO_VALID(minor)) return -ENODEV;
  if(debug) printk(KERN_DEBUG "%s open\n", XADIO_NAME(minor));
  MOD_INC_USE_COUNT;
  return 0;
}

static int xadio_release(struct inode * inode, struct file * file) {
unsigned int minor = MINOR(inode->i_rdev);
  if(debug) printk(KERN_DEBUG "%s release\n", XADIO_NAME(minor));
  MOD_DEC_USE_COUNT;
  return 0;
}

static unsigned int xadio_poll(struct file *file,poll_table *wait) {
unsigned int minor=MINOR(file->f_dentry->d_inode->i_rdev);
  if(debug>1) printk(KERN_DEBUG "%s poll\n", XADIO_NAME(minor));
  poll_wait(file,&(xadio_table[minor].dma_wait),wait);
  switch(xadio_readok(minor,file->f_pos)) {
    case -2:
    case -1:
    case 0:
      return POLLIN | POLLRDNORM;
  }
  return 0;
}

static struct file_operations xadio_fops = {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
  owner:  THIS_MODULE,
#endif
  llseek: xadio_lseek,
  read:   xadio_read,
  write:  xadio_write,
  poll:   xadio_poll,
  open:   xadio_open,
  release:  xadio_release
};

/* --- procfs code   --------------------------------------------------- */

static struct proc_dir_entry *xadioroot;

static void xadio_proc_del_tree(void) {
struct proc_dir_entry *p;
  if(!xadioroot) return;
  for(p=xadioroot->subdir;p;) {
    struct proc_dir_entry *q;
    for(q=p->subdir;q;) {
      struct proc_dir_entry *r=q->next;
      remove_proc_entry(q->name,p);
      q=r;
    }
    q=p->next;
    remove_proc_entry(p->name,xadioroot);
    p=q;
  }
  remove_proc_entry(XADIO_STRING,0);
}

static struct {
  char *name;
  char type;
  int size;
  int flags;
  short offset;
  int space;
} xadio_vars[] = {
  { "00h-LAS0RR",      4,      4,    0664,0x00,0},
  { "04h-LAS0BA",      4,      4,    0664,0x04,0},
  { "08h-MARBR",       4,      4,    0664,0x08,0},
  { "0ch-BIGEND",      1,      1,    0664,0x0c,0},
  { "0dh-LMISC",       1,      1,    0664,0x0d,0},
  { "0eh-PROT_AREA",   2,      2,    0664,0x0e,0},
  { "10h-EROMRR",      4,      4,    0664,0x10,0},
  { "14h-EROMBA",      4,      4,    0664,0x14,0},
  { "18h-LBRD0",       4,      4,    0664,0x18,0},
  { "1ch-DMRR",        4,      4,    0664,0x1c,0},
  { "20h-DMLBAM",      4,      4,    0664,0x20,0},
  { "24h-DMLBAI",      4,      4,    0664,0x24,0},
  { "28h-DMPBAM",      4,      4,    0664,0x28,0},
  { "2ch-DMCFGA",      4,      4,    0664,0x2c,0},
  { "f0h-LAS1RR",      4,      4,    0664,0xf0,0},
  { "f4h-LAS1BA",      4,      4,    0664,0xf4,0},
  { "f8h-LBRD1",       4,      4,    0664,0xf8,0},
  { "fch-DMDAC",       4,      4,    0664,0xfc,0},
  { "80h-DMAMODE0",    4,      4,    0664,0x80,0},
  { "84h-DMAPADR0",    4,      4,    0664,0x84,0},
  { "88h-DMALADR0",    4,      4,    0664,0x88,0},
  { "8ch-DMASIZ0",     4,      4,    0664,0x8c,0},
  { "90h-DMADPR0",     4,      4,    0664,0x90,0},
  { "94h-DMAMODE1",    4,      4,    0664,0x94,0},
  { "98h-DMAPADR1",    4,      4,    0664,0x98,0},
  { "9ch-DMALADR1",    4,      4,    0664,0x9c,0},
  { "a0h-DMASIZ1",     4,      4,    0664,0xa0,0},
  { "a4h-DMADPR1",     4,      4,    0664,0xa4,0},
  { "a8h-DMACSR0",     1,      1,    0664,0xa8,0},
  { "a9h-DMACSR1",     1,      1,    0664,0xa9,0},
  { "ach-DMAARB",      4,      4,    0664,0xac,0},
  { "b0h-DMATHR",      4,      4,    0664,0xb0,0},
  { "b4h-DMADAC0",     4,      4,    0664,0xb4,0},
  { "b8h-DMADAC1",     4,      4,    0664,0xb8,0},
  { "40h-MBOX0",       4,      4,    0664,0x40,0},
  { "44h-MBOX1",       4,      4,    0664,0x44,0},
  { "48h-MBOX2",       4,      4,    0664,0x48,0},
  { "4ch-MBOX3",       4,      4,    0664,0x4c,0},
  { "50h-MBOX4",       4,      4,    0664,0x50,0},
  { "54h-MBOX5",       4,      4,    0664,0x54,0},
  { "58h-MBOX6",       4,      4,    0664,0x58,0},
  { "5ch-MBOX7",       4,      4,    0664,0x5c,0},
  { "60h-P2LDBELL",    4,      4,    0664,0x60,0},
  { "64h-L2PDBELL",    4,      4,    0664,0x64,0},
  { "68h-INTCSR",      4,      4,    0664,0x68,0},
  { "6ch-CNTRL",       4,      4,    0664,0x6c,0},
  { "70h-PCIHIDR",     4,      4,    0444,0x70,0},
  { "74h-PCIHREV",     4,      4,    0444,0x74,0},
  { "78h-MBOX0-QUEUE", 4,      4,    0664,0x78,0},
  { "7ch-MBOX1-QUEUE", 4,      4,    0664,0x7c,0},
  { "30h-OPQIS",       4,      4,    0444,0x30,0},
  { "34h-OPQIM",       4,      4,    0664,0x34,0},
  { "40h-IQP",         4,      4,    0664,0x40,0},
  { "44h-OQP",         4,      4,    0664,0x44,0},
  { "c0h-MQCR",        4,      4,    0664,0xc0,0},
  { "c4h-QBAR",        4,      4,    0664,0xc4,0},
  { "c8h-IFHPR",       4,      4,    0664,0xc8,0},
  { "cch-IFTPR",       4,      4,    0664,0xcc,0},
  { "d0h-IPHPR",       4,      4,    0664,0xd0,0},
  { "d4h-IPTPR",       4,      4,    0664,0xd4,0},
  { "d8h-OFHPR",       4,      4,    0664,0xd8,0},
  { "dch-OFTPR",       4,      4,    0664,0xdc,0},
  { "e0h-OPHPR",       4,      4,    0664,0xe0,0},
  { "e4h-OPTPR",       4,      4,    0664,0xe4,0},
  { "e8h-QSR",         4,      4,    0664,0xe8,0},

  { "00h-ADC0",        4,      4,    0664,0x00,1},
  { "04h-ADC1",        4,      4,    0664,0x04,1},
  { "08h-ADC2",        4,      4,    0664,0x08,1},
  { "0ch-ADC3",        4,      4,    0664,0x0c,1},
  { "10h-ADC4",        4,      4,    0664,0x10,1},
  { "14h-ADC5",        4,      4,    0664,0x14,1},
  { "18h-ADC6",        4,      4,    0664,0x18,1},
  { "1ch-ADC7",        4,      4,    0664,0x1c,1},
  { "20h-DIGIO",       4,      4,    0664,0x20,1},
  { "24h-INCR0",       4,      4,    0664,0x24,1},
  { "28h-DAC0",        4,      4,    0664,0x28,1},
  { "2ch-DAC1",        4,      4,    0664,0x2c,1},
  { "30h-TIMER",       4,      4,    0664,0x30,1},
  { "34h-INCR1",       4,      4,    0664,0x34,1},
  { "38h-VERSION",     4,      4,    0664,0x38,1},
  { "3ch-FILL",        4,      4,    0664,0x3c,1},
  { "40h-SWR",         4,      4,    0664,0x40,1},
  { "44h-IEN",         4,      4,    0664,0x44,1},
  { "48h-IRQ",         4,      4,    0664,0x48,1},
  { "4ch-CONFIG",      4,      4,    0664,0x4c,1},

  { "00h-VENDORID",    2,    2,    0444,0x00,3 },
  { "02h-DEVICEID",    2,    2,    0444,0x02,3 },
  { "04h-PCICR",       2,    2,    0664,0x04,3 },
  { "06h-PCISR",       2,    2,    0664,0x06,3 },
  { "08h-PCIREVID",    1,    1,    0444,0x08,3 },
  { "09h-PCIRLEV",     1,    1,    0444,0x09,3 },
  { "0ah-PCISUBCLS",   1,    1,    0444,0x0a,3 },
  { "0bh-PCIBASCLS",   1,    1,    0444,0x0b,3 },
  { "0ch-PCICLSR",     1,    1,    0664,0x0c,3 },
  { "0dh-PCILTR",      1,    1,    0664,0x0d,3 } ,
  { "0eh-PCIHTR",      1,    1,    0444,0x0e,3 },
  { "0fh-PCIBISTR"    ,1,    1,    0444,0x0f,3 },
  { "10h-PCIBAR0",     4,    4,    0664,0x10,3 },
  { "14h-PCIBAR1",     4,    4,    0664,0x14,3 },
  { "18h-PCIBAR2",     4,    4,    0664,0x18,3 },
  { "1ch-PCIBAR3",     4,    4,    0664,0x1c,3 },
  { "20h-PCIBAR4",     4,    4,    0444,0x20,3 },
  { "24h-PCIBAR5",     4,    4,    0444,0x24,3 },
  { "28h-PCICIS",      4,    4,    0444,0x28,3 },
  { "2ch-PCISVID",     2,    2,    0444,0x2c,3 },
  { "2eh-PCISID",      2,    2,    0444,0x2e,3 },
  { "30h-PCIERBAR",    4,    4,    0664,0x30,3 },
  { "34h-CAP_PTR",     1,    1,    0664,0x34,3 },
  { "3ch-PCIILR",      1,    1,    0664,0x3c,3 },
  { "3dh-PCIIPR",      1,    1,    0444,0x3d,3 },
  { "3eh-PCIMGR",      1,    1,    0444,0x3e,3 },
  { "3fh-PCIMLR",      1,    1,    0444,0x3f,3 },
  { "40h-PMCAPID",     1,    1,    0444,0x40,3 },
  { "41h-PMNEXT",      1,    1,    0664,0x41,3 },
  { "42h-PMC",         2,    2,    0664,0x42,3 },
  { "44h-PMCSR",       2,    2,    0664,0x44,3 },
  { "46h-PMCSR_BSE",   1,    1,    0444,0x46,3 },
  { "47h-PMDATA",      1,    1,    0664,0x47,3 },
  { "48h-HS_CNTL",     1,    1,    0664,0x48,3 },
  { "49h-HS_NEXT",     1,    1,    0664,0x49,3 },
  { "4ah-HS_CSR",      1,    1,    0664,0x4a,3 },
  { "4ch-PVPDCNTL",    1,    1,    0444,0x4c,3 },
  { "4dh-PVPD_NEXT",   1,    1,    0664,0x4d,3 },
  { "4eh-PVPDAD",      2,    2,    0664,0x4e,3 },
  { "50h-PVPDATA",     4,    4,    0664,0x50,3 }
};


static int xadio_getvar(char *buffer,int minor,unsigned int varindex) {
int offset=xadio_vars[varindex].offset;
int type=xadio_vars[varindex].type;
int size=xadio_vars[varindex].size;
int space=xadio_vars[varindex].space;
  return xadio_getvarcore(buffer,minor,offset,type,size,space);
}


static int xadio_setvar(const char *buffer,unsigned int count,int minor,int varindex) {
int offset=xadio_vars[varindex].offset;
int type=xadio_vars[varindex].type;
int size=xadio_vars[varindex].size;
int space=xadio_vars[varindex].space;
  return xadio_setvarcore(buffer,count,minor,offset,type,size,space);
}

static int xadio_vars_write(struct file *file, const char *buffer, unsigned long count, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
unsigned int varindex=(((unsigned int)data) & 0x00ffff00)>>8;
loff_t off=file->f_pos;
  if(debug) printk(KERN_DEBUG "%s vars_write: var:%d off:%Ld count:%ld\n",XADIO_NAME(minor),varindex,off,count);
  return xadio_setvar(buffer,count,minor,varindex);
}

static int xadio_vars_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
unsigned int varindex=(((unsigned int)data) & 0x00ffff00)>>8;
int len;
  if(debug>1) printk(KERN_DEBUG "%s vars_read: var:%d off:%d count:%d\n",XADIO_NAME(minor),varindex,(int)off,count);
  len=xadio_getvar(page,minor,varindex);
  len -= off;
  if (len < count) {
    *eof = 1;
    if (len <= 0) return 0;
  } else {
    len = count;
  }
  *start = page + off;
  return len;
}

/* --------------------- EEprom handling --------------------------------- */

#define PLX_9054_PCI_PVPDAD  0x4e
#define PLX_9054_PCI_PVPDATA 0x50

static int xadio_eeprom_dword_read(int minor,unsigned short offset,u32 *dest) {
int i,j;
u16 eaddr=offset&0x7fff;
struct pci_dev *pdev=xadio_table[minor].pdev;
  *dest=0xbadbadba;
  for(j=0;j<10;j++) {
    if(pci_write_config_word(pdev,PLX_9054_PCI_PVPDAD,eaddr)) {
      printk(KERN_WARNING "%s error writing pci\n",XADIO_NAME(minor));
      config_rom_error[minor] = (config_rom_error[minor]&0xfff0)|0x01;
      return -EIO;
    }
    for(i=0;i<1000;i++) {
      xadio_delay();
      if(signal_pending(current)) return -EINTR;
      if(pci_read_config_word(pdev,PLX_9054_PCI_PVPDAD,&offset)) {
        printk(KERN_WARNING "%s error reading pci\n",XADIO_NAME(minor));
        config_rom_error[minor] = (config_rom_error[minor]&0xfff0)|0x02;
        continue;
      }
      if(offset & 0x8000) {
        if(pci_read_config_dword(pdev,PLX_9054_PCI_PVPDATA,dest)) {
          if(debug)printk(KERN_WARNING "%s error reading pci\n",XADIO_NAME(minor));
          config_rom_error[minor] = (config_rom_error[minor]&0xfff0)|0x03;
          continue;
        }
        config_rom_error[minor] = (config_rom_error[minor]&0xfff0);
        return 0;
      }
    }
  }
  config_rom_error[minor] = (config_rom_error[minor]&0xfff0)|0x04;
  printk(KERN_WARNING "%s EEPROM read error\n",XADIO_NAME(minor));
  return -EIO;
}

static int xadio_eeprom_dword_write(int minor,unsigned short offset,u32 value) {
int i;
struct pci_dev *pdev=xadio_table[minor].pdev;
u16 oldVal=GetPlx_prot(minor);
  SetPlx_prot(minor,0); 
  pci_write_config_dword(pdev,PLX_9054_PCI_PVPDATA,value);
  xadio_sleep(minor,10);
  pci_write_config_word(pdev,PLX_9054_PCI_PVPDAD,offset | 0x8000);
  xadio_sleep(minor,10);
  for(i=0;i<10;i++) {
    pci_read_config_word(pdev,PLX_9054_PCI_PVPDAD,&offset);
    if(!(offset & 0x8000)) break;
    if(signal_pending(current)){
       printk(KERN_WARNING "%s xadio_eeprom_dword_write, signal_pending\n",XADIO_NAME(minor));
       config_rom_error[minor] = (config_rom_error[minor]&0xff0f)|0x10;
       return -EINTR;
    }
    xadio_sleep(minor,100);
  }
  SetPlx_prot(minor,oldVal);
  if(!(offset & 0x8000)){
    config_rom_error[minor] = (config_rom_error[minor]&0xff0f);
    return 0;
  }
  config_rom_error[minor] = (config_rom_error[minor]&0xff0f)|0x20;
  printk(KERN_WARNING "%s xadio_eeprom_dword_write, offset=%04x %04x\n",XADIO_NAME(minor),offset,value);
  return -EAGAIN;
}

static u32 swapu32(u32 w) {
  return
    ((w & 0xff000000)>>24)
   |((w & 0x00ff0000)>>8 )
   |((w & 0x0000ff00)<<8 )
   |((w & 0x000000ff)<<24)
   ;
}

static int xadio_eeprom_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
int size=512;
int i;
  if(debug) printk(KERN_DEBUG "%s eeprom_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  if(count>=size-off) { *eof=1; count=size-off;}
  if(count<=0) { return 0; }
  for(i=0;i<count;i+=4) {
    int error;
    u32 value;
    if((count-i)<4) return i;
    error=xadio_eeprom_dword_read(minor,off+i,&value);
    if(error) return error;
    *((u32*)(page+i))=swapu32(value);
  }
  *start=page;
  return count;
}

static int xadio_eeprom_write_core(int minor,loff_t off,const char *buffer,unsigned long count) {
int size=512;
int i;
char asc[512+1];
  for(i=0; i<(int)count && i<256; i++)
    sprintf(asc+i*2,"%02x",buffer[i]&0xff);
  printk(KERN_DEBUG "%s eeprom_write(%d,%d):.%s.\n",XADIO_NAME(minor),(int)off,(int)count,asc);
  if(count>size-off) { count=size-off; }
  if(count<=0) { return 0; }
  for(i=0;i<(int)count;i+=4) {
    int error;
    u32 value;
    if((count-i)<4) break;
    value=swapu32(*((u32*)(buffer+i)));
    error=xadio_eeprom_dword_write(minor,off+i,value);
    if(error){
       printk(KERN_DEBUG "%s xadio_eeprom_dword_write error=:%d at %d of %ld\n",XADIO_NAME(minor),error,i,count);
       break;
    }
  }
  return i;
}

static int xadio_eeprom_write(struct file *file, const char *buffer, unsigned long count, void *data) {
int minor=((unsigned int)data) & 0xff;
loff_t off=file->f_pos;
  return xadio_eeprom_write_core(minor,off,buffer,count);
}

static int xadio_readconfigrom(int minor) {
char *start;
int eof;
  if(xadio_eeprom_read(xadio_table[minor].configrom, &start,XADIO_CONFIGOFFSET,XADIO_CONFIG_ROMSIZE,&eof,(void*)minor)!=XADIO_CONFIG_ROMSIZE) {
    printk(KERN_WARNING "%s configread failed\n",XADIO_NAME(minor));
    return -EIO;
  }
  if(debug) printk(KERN_WARNING "%s configread ok\n",XADIO_NAME(minor));
  return 0;
}

static int xadio_writeconfigrom(int minor) {
int written,err=0;
  if((written=xadio_eeprom_write_core(minor,XADIO_CONFIGOFFSET,xadio_table[minor].configrom,XADIO_CONFIG_ROMSIZE))!=XADIO_CONFIG_ROMSIZE) {
    printk(KERN_WARNING "%s configwrite failed, written:%d\n",XADIO_NAME(minor),written);
    err= -EIO;
  }
  xadio_readconfigrom(minor);
  xadio_eepromimport(minor);
  xadio_registerfreq(minor,xadio_table[minor].eeprom.curr_freq);
  return err;
}

static int xadio_memory_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
int size2=xadio_table[minor].size2;
  if(debug) printk(KERN_DEBUG "%s memory_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  if(count>=size2-off) { *eof=1; count=size2-off;}
  if(count<=0) { return 0; }
  memcpy_fromio(page,xadio_table[minor].base2+off,count);
  *start=page;
  return count;
}

static int xadio_memory_write(struct file *file, const char *buffer, unsigned long count, void *data) {
int minor=((unsigned int)data) & 0xff;
int size=xadio_table[minor].size2;
loff_t off=file->f_pos;
  if(debug) printk(KERN_DEBUG "%s memory_write: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  if(count>size-off) { count=size-off; }
  if(count<=0) { return 0; }
  memcpy_toio(xadio_table[minor].base2+off,buffer,count);
  return count;
}

static int xadio_config_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
  if(debug) printk(KERN_DEBUG "%s config_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  if(count>=XADIO_CONFIG_ROMSIZE-off) { *eof=1; count=XADIO_CONFIG_ROMSIZE-off;}
  if(count<=0) { return 0; }
  memcpy_fromio(page,xadio_table[minor].base0+off,count);
  *start=page;
  return count;
}

static int xadio_config_write(struct file *file, const char *buffer, unsigned long count, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
loff_t off=file->f_pos;
  if(debug) printk(KERN_DEBUG "%s config_write: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  if(count>XADIO_CONFIG_ROMSIZE-off) { count=XADIO_CONFIG_ROMSIZE-off; }
  if(count<=0) { return 0; }
  memcpy_toio(xadio_table[minor].base0+off,buffer,count);
  return count;
}


/* 0 wenn invalid, 1 wenn 0, sonst normaler wert */
#define ADU2INT(a) (((a) & 0xffff0000)? 0 : ((a)?(a):1))
#define ADU2VAL(a) (((a)>>16)&0x3)
/* Rechnet uV in DACDIGITS um */
/* Rechnet DACDIGITS in uV um */

/* Fixed to +-10.24V */
/*
static int xadio_getchanneloffset(int minor,int channel) {
  return 10240000;
}
*/

/* returns gain * 100  because of precision */
/*
static int xadio_getchannelgain100(int minor,int channel) {
  return 31250;
}
*/

#ifdef LATER
static int xadio_getadc(int minor,int channel) {
  return GetFpga_adu(minor,channel);
}
#endif

static int xadio_adc2value(int minor,int channel,int adc) {
 if((adc & 0xffff0000)==0){
   /*if(adc==0)
     adc=2;*/
   if((xadio_table[minor].eeprom.adccorr[channel].dil & 3) == 3){
     int value0=adc;
     adc = (adc & 0x7fff) | (~adc & 0x8000); /* DAC in bipolarmode*/
     if(debug>0)
       printk(KERN_DEBUG "%s ch[%d] %d -> %d\n",XADIO_NAME(minor),channel,value0,adc);
   }
 }
 else
   return 0;
 return adc;
}

static int xadio_adc2error(int adc) {
  return ((adc)>>16)&0x3;
}
/*
static int xadio_value2uv(int minor,int channel,int value) {
int offset=xadio_getchanneloffset(minor,channel);
int gain=xadio_getchannelgain100(minor,channel);
  return (value*gain)/100-offset;
}
*/
int xadio_value2uv(int minor,int channel,int value) {
int result;
result =  (value-xadio_table[minor].eeprom.adccorr[channel].xm)*xadio_table[minor].eeprom.adccorr[channel].uend/64;
if(debug>=4)
  printk(KERN_DEBUG "%s ch:%d xm:%d uend:%d   %d -> %d\n",XADIO_NAME(minor),channel,xadio_table[minor].eeprom.adccorr[channel].xm,xadio_table[minor].eeprom.adccorr[channel].uend,value,result);
if((xadio_table[minor].eeprom.adccorr[channel].dil & 3) == 2)
  result=-result;
return result;
}

static int xadio_uv2dac(int minor,int channel,int uv) {
int uvo=uv;
  uv=xadio_get_dac_corr(minor,channel,uv);
  if(debug>=4)
    printk(KERN_DEBUG "%s hdio_get_dac_corr: %d -> %d\n",XADIO_NAME(minor),uvo,uv);
  return (((uv<<3)/625)&0x3ffff);
}

static int xadio_dac2uv(int minor,int channel,int dac) {
int result=(dac&0x20000)?((int)(dac|0xfffc0000)):dac;
  return (result*625)>>3;
}

#ifdef LATER
static int xadio_dacuv(int minor,int channel) {
  channel&=1;
  return xadio_dac2uv(minor,channel,GetFpga_dac(minor,channel));
}
#endif

static void xadio_setdacuv(int minor,int channel,int value) {
  channel&=1;
  SetFpga_dac(minor,channel,xadio_uv2dac(minor,channel,value));
}

static void xadio_setdac(int minor,int channel,int value) {
  channel&=1;
  SetFpga_dac(minor,channel,value);
}

static int xadio_gettcelsius(int minor,int channel) 
{
unsigned int mw=xadio_getmodeword(minor);
unsigned int celsius=0;
int          i;
/* nTCS(Bit3) + TIO_EN(Bit2) auf  "1" setzen    TCK(Bit4) auf 0 setzen*/
  mw &= 0xffffffe0; mw |=0x0f; xadio_setmodeword(minor,mw);
/* nTCS(Bit3) auf 0 setzen + wait einlegen */
  mw &=~0x08; xadio_setmodeword(minor,mw);

  for(i=15; i>=0; i--){
    mw |= 0x10; xadio_setmodeword(minor,mw); /* TCK(Bit4) 0->1 Transition */
    mw=xadio_getmodeword(minor);
    if(mw&1)
       celsius |= (1<<i);
    mw &=~0x10; xadio_setmodeword(minor,mw);
  }

/* nTCS(Bit3) + TIO_EN(Bit2) auf  "1" setzen    TCK(Bit4) auf 0 setzen*/
  mw &= 0xffffffe0; mw |=0x0f; xadio_setmodeword(minor,mw);
  return celsius;
}

static int xadio_getversion(int minor,int channel)     { return XADIO_DRIVERVERSION; }
static int xadio_getfpgaversion(int minor,int channel) { return /* 0xadda0210 */ GetFpga_version(minor); } /* not ready */
static int xadio_gethwversion(int minor,int channel)   { return xadio_table[minor].eeprom.hwversion; }
static int xadio_getcardnum(int minor,int channel)     { return xadio_table[minor].eeprom.cardnumber*1000 + xadio_table[minor].eeprom.cardyear; }
static int xadio_getbrecords(int minor,int channel)    { return brecords; }
static int xadio_getbcount(int minor,int channel)      { return bcount; }
static int xadio_getadcbits(int minor,int channel)     { return 16; }
static int xadio_getdacbits(int minor,int channel)     { return 18; }
static int xadio_getchannels(int minor,int channel)    { return XADIO_CHANNELS; }

static int xadio_gettimer32(int minor,int channel) { return GetFpga_timer(minor); }
static int xadio_getinc0(int minor,int channel) { return GetFpga_inc0(minor); }
static int xadio_getinc1(int minor,int channel) { return GetFpga_inc1(minor); }
#ifdef LATER
static int xadio_getdac(int minor,int channel) { return xadio_dacuv(minor,channel); }
#endif
static int xadio_getdigio(int minor,int channel) { return GetFpga_digio(minor); }

static int xadio_getrpm(int minor,int encoder) {
#ifdef USE_RPM_CODE
unsigned int	RpmCounterMode;

	RpmCounterMode=GetFpga_cntrl(minor)&0x20000 ? 1 : 0;
	if(encoder==0 && RpmCounterMode!=0)	return 0;
	if(encoder==1 && RpmCounterMode!=1)	return 0;

	return  GetFpga_fill1(minor);
#else
	return 0;
#endif
}

typedef int (*GetFkt)(int minor,int channel);
typedef int (*SetFkt)(int minor,int channel,int value);

typedef struct {
  int mode;
  int channel;
  const char *name;
  GetFkt   fkt;
} Xadio_FormatEntry;

static Xadio_FormatEntry controlArray[] = {
  { 0, 0, XADIO_CONTROL_version,     xadio_getversion },
  { 0, 0, XADIO_CONTROL_hwversion ,  xadio_gethwversion },
  {13, 0, XADIO_CONTROL_cardnum   ,  xadio_getcardnum },
#ifdef USE_DEBUG_IN_CONTROL
  {14, 0, XADIO_CONTROL_debug     ,  (GetFkt)0 },
#endif
  { 1, 0, XADIO_CONTROL_fpgaversion, xadio_getfpgaversion   },
  { 0, 0, XADIO_CONTROL_brecords ,   xadio_getbrecords },
  { 9, 0, XADIO_CONTROL_tcelsius,    xadio_gettcelsius },
  { 0, 0, XADIO_CONTROL_bcount ,     xadio_getbcount },
  { 0, 0, XADIO_CONTROL_freq ,       xadio_getfreq },
  { 0, 0, XADIO_CONTROL_incmode0,    xadio_getincmode },
  { 0, 1, XADIO_CONTROL_incmode1,    xadio_getincmode },
  {12, 0, XADIO_CONTROL_div0,        xadio_getdiv },
  {12, 1, XADIO_CONTROL_div1,        xadio_getdiv },
  { 0, 0, XADIO_CONTROL_adcbits,     xadio_getadcbits },
  { 0, 0, XADIO_CONTROL_dacbits,     xadio_getdacbits },
  { 0, 0, XADIO_CONTROL_channels,    xadio_getchannels },
/*
  { 0, 0, XADIO_CONTROL_EEversion,   xadio_geteepromversion },
  { 0, 0, XADIO_CONTROL_EEhwversion, xadio_geteepromhardwarerevision },
  { 0, 0, XADIO_CONTROL_EEcardyear,  xadio_geteepromcardyear },
  { 0, 0, XADIO_CONTROL_EEcardnumber,xadio_geteepromcardnumber },
  { 1, 0, XADIO_CONTROL_EEjumper0,   xadio_geteepromjumpers },
  { 1, 1, XADIO_CONTROL_EEjumper1,   xadio_geteepromjumpers },
  { 1, 2, XADIO_CONTROL_EEjumper2,   xadio_geteepromjumpers },
  { 1, 3, XADIO_CONTROL_EEjumper3,   xadio_geteepromjumpers },
  { 1, 4, XADIO_CONTROL_EEjumper4,   xadio_geteepromjumpers },
  { 1, 5, XADIO_CONTROL_EEjumper5,   xadio_geteepromjumpers },
  { 1, 6, XADIO_CONTROL_EEjumper6,   xadio_geteepromjumpers },
  { 1, 7, XADIO_CONTROL_EEjumper7,   xadio_geteepromjumpers },
  { 8, 0, XADIO_CONTROL_EEcalib0,    (GetFkt)0 },
  { 8, 1, XADIO_CONTROL_EEcalib1,    (GetFkt)0 },
  { 8, 2, XADIO_CONTROL_EEcalib2,    (GetFkt)0 },
  { 8, 3, XADIO_CONTROL_EEcalib3,    (GetFkt)0 },
  { 8, 4, XADIO_CONTROL_EEcalib4,    (GetFkt)0 },
  { 8, 5, XADIO_CONTROL_EEcalib5,    (GetFkt)0 },
  { 8, 6, XADIO_CONTROL_EEcalib6,    (GetFkt)0 },
  { 8, 7, XADIO_CONTROL_EEcalib7,    (GetFkt)0 },
*/
  { 6, 0, XADIO_CONTROL_frequencies, (GetFkt)0 },
  { 7, 0, XADIO_CONTROL_configrom,   (GetFkt)0 }
};

static Xadio_FormatEntry valuesArray[] = {
  { 2, 0, XADIO_VALUES_timer,       (GetFkt)0         },
  { 10,0, XADIO_VALUES_timer32,     xadio_gettimer32  },
  { 0, 0, XADIO_VALUES_inc0,        xadio_getinc0     },
  { 0, 0, XADIO_VALUES_inc1,        xadio_getinc1     },
  { 4, 0, XADIO_VALUES_adc0 ,       (GetFkt)0         },
  { 4, 1, XADIO_VALUES_adc1 ,       (GetFkt)0         },
  { 4, 2, XADIO_VALUES_adc2 ,       (GetFkt)0         },
  { 4, 3, XADIO_VALUES_adc3 ,       (GetFkt)0         },
  { 4, 4, XADIO_VALUES_adc4 ,       (GetFkt)0         },
  { 4, 5, XADIO_VALUES_adc5 ,       (GetFkt)0         },
  { 4, 6, XADIO_VALUES_adc6 ,       (GetFkt)0         },
  { 4, 7, XADIO_VALUES_adc7 ,       (GetFkt)0         },
  { 5, 0, XADIO_VALUES_dac0,        (GetFkt)0         },
  { 5, 1, XADIO_VALUES_dac1,        (GetFkt)0         },
  { 3, 0, XADIO_VALUES_digio,       xadio_getdigio    },
  { 11,0, XADIO_VALUES_rpm0,        (GetFkt)0         },
  { 11,1, XADIO_VALUES_rpm1,        (GetFkt)0         }
};

static int xadio_bool(unsigned int value,int bits,char *buffer,int size) {
int i;
int pos=0;
  if(size<bits+bits/4+2) return 0;
  for(i=0;i<bits;i++) {
    if(!(i%4)) buffer[pos++]='.';
    buffer[pos++]=(value&(0x1<<(bits-i-1)))?'1':'0';
  }
  buffer[pos++]=0;
  return pos;
}

static int xadio_format(int minor,char *buffer,int size,Xadio_FormatEntry *fe,int asize) {
int div,numyear,count=0,timer32=0,use_same_timer=-1/* -1 or 0 */;
int i;
  for(i=0;i<asize;i++) {
    if(count>size-50) break;
    switch(fe[i].mode) {
      case 0:
        count+=sprintf(buffer+count,"%s %d\n",fe[i].name,(fe[i].fkt)(minor,fe[i].channel));
        break;
      case 1:
        count+=sprintf(buffer+count,"%s 0x%x\n",fe[i].name,(fe[i].fkt)(minor,fe[i].channel));
        break;
      case 2: 
        timer32=xadio_gettimer32(minor,0); 
        count+=sprintf(buffer+count,"%s 0x%8.8x%8.8x\n",fe[i].name,HILO(xadio_corrpos(minor,timer32)>>6));
        if(use_same_timer==-1)
           use_same_timer=1;
        break;
      case 3: {
        char lbuffer[100];
        int value=(fe[i].fkt)(minor,fe[i].channel);
        xadio_bool(value,24,lbuffer,sizeof(lbuffer));
        count+=sprintf(buffer+count,"%s 0x%6.6x 0b%s\n",fe[i].name,value,lbuffer);
        break;
      }
      case 4: {
        int channel=fe[i].channel;
        int adc=GetFpga_adu(minor,channel);
        int value=xadio_adc2value(minor,channel,adc);
        int uv=xadio_value2uv(minor,channel,xadio_adccorr(minor,channel,value));
        int error=xadio_adc2error(adc);
        count+=sprintf(buffer+count,"%s 0x%6.6x % 9d %d\n",fe[i].name,value,uv,error);
        break;
      }
      case 5: {
        int channel=fe[i].channel;
        int dac=GetFpga_dac(minor,channel);
        int uv=xadio_dac2uv(minor,channel,dac);
        count+=sprintf(buffer+count,"%s 0x%6.6x % 9d\n",fe[i].name,dac,uv);
        break;
      }
      case 6: {
        int j;
        count+=sprintf(buffer+count,"%s {",fe[i].name);
        for(j=0;j<128;j++) {
          int frequency=xadio_rate2freq(j);
          if(frequency>195312 || frequency<3051) continue; /* AD7222 Anschlag */
          if(count>size-50) continue;
          count+=sprintf(buffer+count,"%d ",frequency);
        }
        count+=sprintf(buffer+count,"}\n");
        break;
      }
      case 7: {
        int j;
        xadio_readconfigrom(minor);
        count+=sprintf(buffer+count,"%s ",fe[i].name);
        for(j=0;j<XADIO_CONFIG_ROMSIZE;j++) {
          if(count>size-50) continue;
          count+=sprintf(buffer+count,"%02x",((unsigned char*)(xadio_table[minor].configrom))[j]);
        }
        count+=sprintf(buffer+count,"\n"); 
        break;
      }
/*
      case 8: {
        int j;
        int channel=fe[i].channel;
        count+=sprintf(buffer+count,"%s ",fe[i].name);
        for(j=0;j<xadio_table[minor].eeprom.entries;j++) {
          if(count>size-50) continue;
          count+=sprintf(buffer+count,"{ %d:%d %d }",
             xadio_table[minor].eeprom.entry[j].frequency,
             xadio_table[minor].eeprom.entry[j].corr0.offsets[channel],
             xadio_table[minor].eeprom.entry[j].corr1.offsets[channel]
            );
        }
        count+=sprintf(buffer+count,"\n");
        break;
      }
*/
      case 9: {
        int celsius=(fe[i].fkt)(minor,fe[i].channel);
        count+=sprintf(buffer+count,"%s 0x%04x %d.%02d %d",fe[i].name,celsius,(celsius>>3)/16,((celsius>>3)%16)*100/16,debug); /* debug-mode */
        count+=sprintf(buffer+count,"\n");
        break;
      }
      case 10:
        if(use_same_timer==1)
          count+=sprintf(buffer+count,"%s %u\n",fe[i].name,(0xffffffff)&timer32); /* 0x7fffffff: still 31 Bit output to be compatible with older applications */
        else
          count+=sprintf(buffer+count,"%s %u\n",fe[i].name,(0xffffffff)&(fe[i].fkt)(minor,fe[i].channel)); /* 0x7fffffff: still 31 Bit output to be compatible with older applications */
        break;
      case 11: {
	    unsigned int counter,rpma=0,rpmb=0;
        unsigned int encoder = fe[i].channel;
	    unsigned int div     = encoder==0 ? GetFpga_div0(minor)  : GetFpga_div1(minor);
	    unsigned int f       = 25000000;
        const    int edges   = 4;   /*4edges=1increment*/
        unsigned int encodersize0=xadioencodersize[minor][encoder];
        unsigned int encodersize=encodersize0;
        char         simplerpm[128];
        
        counter=xadio_getrpm(minor,encoder);
        f*=(60*4)/edges;
	    encodersize*=(div+1);
	    /* reduce overflow risk */
	    while( (f&1)==0 && (encodersize&1)==0 ){
		    f>>=1;
		    encodersize>>=1;
	    }

	    if(counter!=0){
		    /*
		    rpm0   11.73 5859375 31199 16
		    rpm0   11.73 5859375 31215 16
		    rpm0   12.50 5859375 29279 16
		    */
            unsigned ocount=counter*2*(div+1);
            unsigned ocount_per_sec=ocount*encodersize0;
            simplerpm[0]=0;
            sprintf(simplerpm,"ocount:%u ocount_per_sec:%u rps:%u rpm:%u", ocount,ocount_per_sec,50000000/ocount_per_sec,50000000U*60U/ocount_per_sec);
		    rpma=f/counter/encodersize;
			/* rpm0 36621.09 f:5859375 counter:40 encodersize:4 ocount:80 ocount_per_sec:81920 rps:610 rpm:36621 */
		    if(f<42949672){
			    rpmb=100*f/counter/encodersize;
        	    count+=sprintf(buffer+count,"%s %4u.%02u %u %u %u %s \n",fe[i].name,rpma,rpmb-100*rpma,f,counter,encodersize,simplerpm);
		    } else	if(f<429496729){
			    rpmb=10*f/counter/encodersize;
        	    count+=sprintf(buffer+count,"%s %4u.%01u  %u %u %u %s\n",fe[i].name,rpma,rpmb-100*rpma,f,counter,encodersize,simplerpm);
		    }
		    else{
        	    count+=sprintf(buffer+count,"%s %4u   %u %u %u %s\n",fe[i].name,rpma,f,counter,encodersize,simplerpm);
		    }
	    }
	    else
        	/* count+=sprintf(buffer+count,"%s    -    %u %u %u\n",fe[i].name,f,counter,encodersize); */
        	count+=sprintf(buffer+count,"%s    -\n",fe[i].name);
        break;
        }
      case 12:{
	    int encoder=fe[i].channel;
        div = (fe[i].fkt)(minor,encoder);
	    if(div!=0)
	  	    printk(KERN_WARNING "%s control div%d\n",XADIO_NAME(minor),div);
        count+=sprintf(buffer+count,"%s %d %d",fe[i].name,div,xadioencodersize[minor][encoder]); /* debug-mode */
        count+=sprintf(buffer+count,"\n");
        break;
        }		
      case 13:
        numyear=(fe[i].fkt)(minor,fe[i].channel);
        count+=sprintf(buffer+count,"%s %02d/%02d\n",fe[i].name,numyear/1000,numyear%1000);
        break;
#ifdef USE_DEBUG_IN_CONTROL
      case 14:
        count+=sprintf(buffer+count,"%s %d %d %s\n",fe[i].name,xadio_table[minor].debug_interrupts,xadio_table[minor].aktbuf,xadio_table[minor].debugtext);
        xadio_table[minor].debugtext[0]=0;
        break;
#endif
    }
  }
  buffer[count]=0;
  return count;
}


static int xadio_common_format(int minor,char *page, char **start, off_t off, int count, int *eof, void *data,Xadio_FormatEntry *fe,int asize) {
int size;
  *eof=1;
  if(off!=0) { return 0; }
  size=xadio_format(minor,page,count,fe,asize);
  *start=page;
  return size;
}

static int xadio_control_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
  if(debug) printk(KERN_DEBUG "%s xadio_control_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  return xadio_common_format(minor,page,start,off,count,eof,data,controlArray,ARRSIZE(controlArray));
}

static int xadio_values_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
int minor=((unsigned int)data) & 0xff;
  if(debug>=4) printk(KERN_DEBUG "%s xadio_values_read: off:%d count:%d\n",XADIO_NAME(minor),(int)off,(int)count);
  return xadio_common_format(minor,page,start,off,count,eof,data,valuesArray,ARRSIZE(valuesArray));
}


static void control_interpreter(int minor,const char *buffer,int len) {
#define TN(x) (len>=(int)strlen(x) && !strncmp(buffer,x,strlen(x))) 
  if(!len) return;
  if(TN(XADIO_CONTROL_setfreq)) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(xadio_setfreq(minor,value)) goto error;
  } else if(TN("setincmode0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>3) goto error;
    xadio_setincmode0(minor,value);
  } else if(TN("setincmode1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>3) goto error;
    xadio_setincmode1(minor,value);
  } else if(TN("setrpmencoder")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    printk(KERN_WARNING "%s setrpmencoder i=%d buffer[i]=%s\n",XADIO_NAME(minor),i,buffer+i);
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>1) goto error;
    xadio_setrpmencoder(minor,value);
  } else if(TN("setdiv0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0) goto error;
    for(   ;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&xadioencodersize[minor][0])) goto error;
    xadio_setdiv0(minor,value);
  } else if(TN("setdiv1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0) goto error;
    for(   ;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&xadioencodersize[minor][1])) goto error;
    xadio_setdiv1(minor,value);
  } else if(TN("resetcounter")) {
    xadio_resetcounter(minor);
  } else if(TN("calibrate")) {
    xadio_calibrate(minor);
  } else if(TN("resetadu")) {
    xadio_resetadu(minor);
  } else if(TN("resetint")) {
    xadio_resetint(minor);
  } else if(TN("dmastop")) {
    xadio_dmastop(minor);
  } else if(TN("dmastart")) {
    xadio_dmastart(minor);
  } else if(TN("reset")) {
    if(xadio_reset(minor)) {
      printk(KERN_WARNING "%s reset failed\n",XADIO_NAME(minor));
    } else {
      xadio_greenled(minor,1);
    }
  } else if(TN("settestmode")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<0 || value>3) goto error;
    xadio_settestmode(minor,value);
  } else if(TN("setdebugmode")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    debug=value;
  } else if(TN("setdigio")) {
    int i,value,mask;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    for(   ;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&mask)) {
      mask=0xfffffff;
    }
    xadio_setdigio(minor,value,mask);
  } else if(TN("setdax0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    xadio_setdac(minor,0,value);
  } else if(TN("setdax1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    xadio_setdac(minor,1,value);
  } else if(TN("setda0")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<-10240000 || value>10240000) goto error;
    xadio_setdacuv(minor,0,value);
  } else if(TN("setda1")) {
    int i,value;
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanint(&buffer[i],len-i,&value)) goto error;
    if(value<-10240000 || value>10240000) goto error;
    xadio_setdacuv(minor,1,value);
  } else if(TN(XADIO_CONTROL_setconfigrom)) {
    int i,res;
    char lbuffer[1024];
    for(i=0;i<len;i++) if(isspace(buffer[i])) break;
    for(   ;i<len;i++) if(!isspace(buffer[i])) break;
    if(xadio_scanhex(&buffer[i],len-i,lbuffer,sizeof(lbuffer),&res)) goto error;
    memcpy(xadio_table[minor].configrom,lbuffer,res);
    xadio_writeconfigrom(minor);
  } else {
    goto error;
  }
  return;
error:
  printk(KERN_WARNING "%s error in control device '%.20s`\n",XADIO_NAME(minor), buffer);
#undef TN
}

static int xadio_control_write(struct file *file, const char *buffer, unsigned long count, void *data) {
unsigned int minor=((unsigned int)data) & 0xff;
loff_t off=file->f_pos;
int lstart=0;
int i;
  if(debug>1) printk(KERN_DEBUG "%s control_write: off:%d count:%d value:%s",XADIO_NAME(minor),(int)off,(int)count,buffer);
  for(i=0;i<(int)count;i++) {
    if(buffer[i]=='\n') {
      control_interpreter(minor,buffer+lstart,i-lstart);
      lstart=i+1;
    }
  }
  if(lstart!=(int)count) {
    control_interpreter(minor,buffer+lstart,i-lstart);
  }
  return count;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))

static struct inode_operations xadio_inode_operations = {
  &xadio_fops,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

static void xadio_fill_inode(struct inode * inode, int fill) {
int minor=(int)(inode->i_size);
  if(!XADIO_VALID(minor)) return;
  inode->i_rdev=MKDEV(major,minor);
  inode->i_op=&xadio_inode_operations;
}
#endif

static int xadio_proc_register(unsigned int minor) {
struct proc_dir_entry *card,*entry,*regs;
int i;
  if(!xadioroot) {
    if(!(xadioroot=create_proc_entry(XADIO_STRING ,S_IFDIR|0555,0))) return -ENOSPC;
  }
  if(!(card=create_proc_entry(xadio_table[minor].name,S_IFDIR|0555,xadioroot))) return -ENOSPC;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
  if(!(entry=create_proc_entry("samples",S_IFCHR|0666,card))) return -ENOSPC;
  entry->size=minor;
  entry->fill_inode=xadio_fill_inode;
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  if(!(entry = proc_mknod("samples",S_IFCHR|0666,card,MKDEV(major,minor)))) return -ENOSPC;
#else
  if(!(entry=create_proc_entry("samples",S_IFCHR|0666,card))) return -ENOSPC;
  entry->proc_fops=&xadio_fops;
  entry->data=(void*)(minor&0xff);
#endif
  if(!(entry=create_proc_entry("control",S_IFREG|0666,card))) return -ENOSPC;
  entry->read_proc=xadio_control_read;
  entry->write_proc=xadio_control_write;
  entry->data=(void*)(minor & 0xff);
  if(!(entry=create_proc_entry("values",S_IFREG|0666,card))) return -ENOSPC;
  entry->read_proc=xadio_values_read;
  entry->write_proc=xadio_control_write;
  entry->data=(void*)(minor & 0xff);
  if(debug) {
    if(!(regs=create_proc_entry("plx",S_IFDIR|0555,card))) return -ENOSPC;
    for(i=0;i<ARRSIZE(xadio_vars);i++) {
      if((xadio_vars[i].space==0) || (xadio_vars[i].space==2) ) {
        entry=create_proc_entry(xadio_vars[i].name,xadio_vars[i].flags,regs);
        if(!entry) return -ENOSPC;
        entry->read_proc=xadio_vars_read;
        entry->write_proc=xadio_vars_write;
        entry->size=xadio_vars[i].size;
        entry->data=(void*)((minor & 0xff)+((i & 0xffff)<<8));
      }
    }
  
    if(!(regs=create_proc_entry("fpga",S_IFDIR|0555,card))) return -ENOSPC;
    for(i=0;i<ARRSIZE(xadio_vars);i++) {
      if(xadio_vars[i].space==1) {
        entry=create_proc_entry(xadio_vars[i].name,xadio_vars[i].flags,regs);
        if(!entry) return -ENOSPC;
        entry->read_proc=xadio_vars_read;
        entry->write_proc=xadio_vars_write;
        entry->size=xadio_vars[i].size;
        entry->data=(void*)((minor & 0xff)+((i & 0xffff)<<8));
      }
    }

    if(!(regs=create_proc_entry("pci",S_IFDIR|0555,card))) return -ENOSPC;
    for(i=0;i<ARRSIZE(xadio_vars);i++) {
      if(xadio_vars[i].space==3) {
        entry=create_proc_entry(xadio_vars[i].name,xadio_vars[i].flags,regs);
        if(!entry) return -ENOSPC;
        entry->read_proc=xadio_vars_read;
        entry->write_proc=xadio_vars_write;
        entry->size=xadio_vars[i].size;
        entry->data=(void*)((minor & 0xff)+((i & 0xffff)<<8));
      }
    }
 
    if(debug>2) { 
      entry=create_proc_entry("config",S_IFREG|0666,card);
      if(!entry) return -ENOSPC;
      entry->read_proc=xadio_config_read;
      entry->write_proc=xadio_config_write;
      entry->size=XADIO_CONFIG_ROMSIZE;
      entry->data=(void*)(minor&0xff);
    }

    entry=create_proc_entry("eeprom",S_IFREG|0666,card);
    if(!entry) return -ENOSPC;
    entry->read_proc=xadio_eeprom_read;
    entry->write_proc=xadio_eeprom_write;
    entry->size=512;
    entry->data=(void*)(minor&0xff);

    entry=create_proc_entry("target",S_IFREG|0644,card);
    if(!entry) return -ENOSPC;
    entry->read_proc=xadio_memory_read;
    entry->write_proc=xadio_memory_write;
    entry->size=xadio_table[minor].size2;
    entry->data=(void*)(minor&0xff);
  }
  return 0;
}

/* ------------------- Card-Detection and Initializing  ------------------ */

static void __devexit xadio_remove_card(int minor) {
int j;
struct pci_dev *pdev=xadio_table[minor].pdev;
  if(debug>1) printk(KERN_DEBUG "%s remove_card\n",XADIO_NAME(minor));
  xadio_dmastop(minor);
  xadio_sleep(minor,50);
  xadio_plxintmode(minor,0);
  xadio_table[minor].flags&=~(XADIO_VALID_MASK);
  free_irq(pdev->irq,&(xadio_table[minor]));
  xadio_cpldled(minor,0);  // LED off
  xadio_greenled(minor,0);
  pci_set_drvdata(pdev,0);
  if(xadio_table[minor].base2) {
    SetFpga_ien(minor,0x00000000);  /* Disable all interrupts */
  }
  if(xadio_table[minor].base0) iounmap((void*)xadio_table[minor].base0);
  xadio_table[minor].base0=0;
  if(xadio_table[minor].base2) iounmap((void*)(xadio_table[minor].base2));
  xadio_table[minor].base2=0;
  for(j=0;j<bcount;j++) {
    if(xadio_table[minor].barr[j]) kfree((void*)(xadio_table[minor].barr[j]));
    xadio_table[minor].barr[j]=0;
    if(xadio_table[minor].darr[j]) kfree((void*)(xadio_table[minor].darr[j]));
    xadio_table[minor].darr[j]=0;
  }
  pci_release_regions(pdev);
  pci_disable_device(pdev);
  xadio_table[minor].flags&=~(XADIO_USED_MASK);
  memset(&xadio_table[minor],0,sizeof(xadio_table[minor])); //shureisshure
}

static int __devinit xadio_getFreeMinor(void) {
int i;
  for(i=0;i<XADIO_NO;i++) {
    if(!XADIO_USED(i)) {
      xadio_table[i].flags|=XADIO_USED_MASK;
      return i;
    }
  }
  return -ENOSPC;
}

static void __devexit xadio_remove(struct pci_dev *pdev) {
int minor=(int)pci_get_drvdata(pdev)-1;
  if(debug>1) printk(KERN_DEBUG "%s xadio_remove(%d)\n",XADIO_NAME(minor),minor);
  xadio_remove_card(minor);
}

static int __devinit xadio_probe(struct pci_dev *pdev,const struct pci_device_id *entry) {
int i;
int error=0;
int irq = pdev->irq;
unsigned long base0=pci_resource_start(pdev,0);
unsigned long base2=pci_resource_start(pdev,2);
unsigned long size0=pci_resource_len(pdev,0);
unsigned long size2=pci_resource_len(pdev,2);
int minor;
  printk(KERN_DEBUG XADIO_STRING " card dectected IRQ:%d Base0:0x%lx Size0:0x%lx Base2:0x%lx Size2:0x%lx\n",irq,base0,size0,base2,size2);
  minor=xadio_getFreeMinor();
  if(minor<0) return -ENOSPC;
  xadio_table[minor].pdev=pdev;
  sprintf(xadio_table[minor].name,XADIO_STRING "%d",minor);
  if(pci_enable_device(pdev)) {
    printk(KERN_WARNING "%s pci_enable_device failed\n",XADIO_NAME(minor));
    error++;
  }
  pci_set_master(pdev);
  if(pci_request_regions(pdev,XADIO_STRING)) {
    printk(KERN_WARNING "%s pci_request_regions failed\n",XADIO_NAME(minor));
    error++;
  }
  if(!(xadio_table[minor].base0=(volatile unsigned char *)ioremap(base0,size0))) {
    printk(KERN_WARNING "%s ioremap failed:base0\n",XADIO_NAME(minor));
    error++;
  }
  xadio_table[minor].size0=size0;
  if(size2) {
    if(!(xadio_table[minor].base2=(volatile unsigned char *)ioremap(base2,size2)))  {
      printk(KERN_WARNING "%s ioremap failed:base2\n",xadio_table[minor].name);
      error++;
    }
    xadio_table[minor].size2=size2;
  }
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0))
  init_waitqueue_head(&(xadio_table[minor].dma_wait));
  init_waitqueue_head(&(xadio_table[minor].local_wait));
  init_waitqueue_head(&(xadio_table[minor].write_wait));
#endif
  for(i=0;i<bcount;i++) {
    if(!(xadio_table[minor].barr[i]=kmalloc(xadio_roundmemsize16(bsize),GFP_KERNEL))) {
      printk(KERN_WARNING "%s kmalloc failed:buffer\n",xadio_table[minor].name);
      error++;
    }
    if(!(xadio_table[minor].darr[i]=kmalloc(xadio_roundmemsize16(sizeof(xadio_dma_desc)),GFP_KERNEL))) {
      printk(KERN_WARNING "%s kmalloc failed:desc\n", xadio_table[minor].name);
      error++;
    }
    xadio_table[minor].darr[i]->dst=virt_to_bus(
       xadio_roundptr16(xadio_table[minor].barr[i])
    );
    if(debug>2) {
      printk(KERN_DEBUG "%s dst:[%d]=%8.8x\n", xadio_table[minor].name,i,xadio_table[minor].darr[i]->dst);
    }
    xadio_table[minor].darr[i]->src=0x02000000;
    xadio_table[minor].darr[i]->size=bsize;
  }
  for(i=0;i<bcount;i++) {
    xadio_table[minor].darr[i]->next=virt_to_bus(
      xadio_roundptr16(xadio_table[minor].darr[(i+1)%bcount])
      )|0xd;
    if(debug>2) {
      printk(KERN_DEBUG "%s next:[%d]=%8.8x\n", xadio_table[minor].name,i,xadio_table[minor].darr[i]->next);
    }
  }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
  if(request_irq(pdev->irq,xadio_interrupt,SA_SHIRQ,xadio_table[minor].name,&(xadio_table[minor]))) {
#else
  if(request_irq(pdev->irq,xadio_interrupt,IRQF_SHARED,xadio_table[minor].name,&(xadio_table[minor]))) {
#endif
    printk(KERN_INFO "%s no irq -> polling used\n", xadio_table[minor].name);
  }
  if(error) {
    xadio_remove_card(minor);
    return -EIO;
  }
  if(xadio_reset(minor)) {
    xadio_remove_card(minor);
    return -EIO;
  }
  xadio_readconfigrom(minor);
  xadio_eepromimport(minor);
  /* again to call xadio_registerfreq after xadio_eepromimport */
  xadio_sleep(minor,20);
  xadio_setfreq(minor,195312);
  xadio_calibrate(minor); /* makes very first values valid */
  if(xadio_proc_register(minor)) {
    printk(KERN_WARNING "%s Error in proc_register\n",xadio_table[minor].name);
  }
  xadio_setdacuv(minor,0,0);
  xadio_table[minor].flags |= XADIO_VALID_MASK;
  xadio_cpldled(minor,1); // LED on
  xadio_greenled(minor,1);
  pci_set_drvdata(pdev,(void*)(minor+1));
  return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static struct pci_device_id xadio_pci_tbl_normal[] = {
  {PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_9054, PCI_VENDOR_ID_DRESEARCH, PCI_DEVICE_ID_XADIO},
  {0}                    /* End of list */
};

static struct pci_device_id xadio_pci_tbl_nosubvendorcheck[] = {
  {PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_9054, PCI_ANY_ID, PCI_ANY_ID},
  {0}                    /* End of list */
};

static struct pci_driver xadio_driver = {
  name:"xadio",
  id_table:xadio_pci_tbl_normal,
  probe:xadio_probe,
  remove:xadio_remove,
};

static int __init xadio_init(void) {
  bsize=brecords*sizeof(Xadio_record);
  if(nosubvendorcheck) {
    xadio_driver.id_table=xadio_pci_tbl_nosubvendorcheck;
  }
  /* if(!pci_present()) return -ENODEV; */
  if(pci_register_driver(&xadio_driver)<0) {
    printk(KERN_INFO XADIO_STRING ":no card found\n");
    pci_unregister_driver(&xadio_driver);
    return -ENODEV;
  }
  register_chrdev(major, XADIO_STRING ,&xadio_fops);
  printk(KERN_INFO XADIO_STRING " driver loaded V%d, debug=%d\n",XADIO_DRIVERVERSION,debug);
  return 0;
}

static void __exit xadio_exit(void) {
  unregister_chrdev(major, XADIO_STRING);
  xadio_proc_del_tree();
  pci_unregister_driver(&xadio_driver);
  printk(KERN_INFO XADIO_STRING " driver unloaded\n");
}
#else

static int __init xadio_init(void) {
int loaded=0;
struct pci_dev *pdev=0;
  bsize=brecords*sizeof(Xadio_record);
  if(!pci_present()) return -ENODEV;
  if(debug) printk(KERN_INFO XADIO_STRING " driver starting\n");
  while((pdev=pci_find_device(PCI_VENDOR_ID_PLX,PCI_DEVICE_ID_9054,pdev))) {
    u16 subvendor;
    pci_read_config_word(pdev,PCI_SUBSYSTEM_VENDOR_ID,&subvendor);
    if(subvendor==PCI_VENDOR_ID_DRESEARCH) {
      if(debug) printk(KERN_INFO XADIO_STRING " card probing\n");
      if(!xadio_probe(pdev,0)) loaded++;
    }
  }
  if(loaded) {
    if(debug) printk(KERN_INFO XADIO_STRING " driver loaded\n");
    return 0;
  } else {
    if(debug) printk(KERN_INFO XADIO_STRING " no card found\n");
    return -ENODEV;
  }
}

static void xadio_exit(void) {
int i;
  unregister_chrdev(major, XADIO_STRING);
  for(i=0;i<XADIO_NO;i++) {
    if(xadio_table[i].pdev) {
      xadio_remove(xadio_table[i].pdev);
    }
  }
  xadio_proc_del_tree();
  printk(KERN_INFO XADIO_STRING " driver unloaded\n");
}

#endif

/* ------------------- Module-Code  ------------------ */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
MODULE_PARM(debug,"i");
MODULE_PARM(major,"i");
MODULE_PARM(brecords,"i");
MODULE_PARM(bcount,"i");
MODULE_PARM(nosubvendorcheck,"i");
MODULE_PARM(hwversion,"i");
#else
module_param(debug,int,0600);
module_param(major,int,0600);
module_param(brecords,int,0600);
module_param(bcount,int,0600);
module_param(nosubvendorcheck,int,0600);
module_param(hwversion,int,0600);
#endif

MODULE_PARM_DESC(debug,"Debuglevel (default:0)");
MODULE_PARM_DESC(major,"Major device number (default:0x46/70)");
MODULE_PARM_DESC(brecords,"Buffersize in Samples (64 Bytes) (default:"__MODULE_STRING(XADIO_DEFAULT_BRECORDS)" Records)");
MODULE_PARM_DESC(bcount,"Buffercount per card 2..16(default:" __MODULE_STRING(XADIO_DEFAULT_BCOUNT) ")");
MODULE_PARM_DESC(nosubvendorcheck,"Dont Check Subvendorid(default:" __MODULE_STRING(XADIO_DEFAULT_NOSUBVENDORCHECK) ")");
MODULE_PARM_DESC(hwversion,"Hardwareversion(default:" __MODULE_STRING(XADIO_DEFAULT_HWVERSION) ")");

MODULE_AUTHOR("Peter Brueckner (peter@bj-ig.de)");
MODULE_DESCRIPTION("Driver for XADIO Cards");
MODULE_SUPPORTED_DEVICE("XADIO XTREME AnalogDigitalIO 8");
MODULE_LICENSE("GPL");

module_init(xadio_init);
module_exit(xadio_exit);
